# -*- coding: utf-8 -*-
import os
import responses
import json
import pytest
import requests
import multipart
import urllib.parse
from io import BytesIO

from zgrobot import ZgRoBot
from zgrobot.config import Config
from zgrobot.client import Client, check_error, ClientException
from zgrobot.utils import cached_property

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
GOD_PIC = os.path.join(os.path.dirname(__file__), '照桥心美.png')
TOKEN_URL = "https://api.weixin.qq.com/cgi-bin/token"
JSON_HEADER = {'content-type': 'application/json'}


def token_callback(request):
    return 200, JSON_HEADER, json.dumps(
        {
            "access_token": "ACCESS_TOKEN",
            "expires_in": 7200
        }
    )


def add_token_response(method):
    def wrapped_func(self, *args, **kwargs):
        responses.add_callback(
            responses.GET, TOKEN_URL, callback=token_callback
        )
        method(self, *args, **kwargs)

    return wrapped_func


def create_pic_file(func):
    def wrapped_func(self, *args, **kwargs):
        with open(GOD_PIC, 'a') as f:
            f.write("just a test")
        try:
            func(self, *args, **kwargs)
        finally:
            os.remove(GOD_PIC)

    return wrapped_func


class BaseTestClass:
    @cached_property
    def client(self):
        config = Config()
        config.from_pyfile(os.path.join(BASE_DIR, "client_config.py"))
        return Client(config)

    @staticmethod
    def callback_without_check(request):
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})


class TestClientBaseClass(BaseTestClass):
    def test_id_and_secret(self):
        assert self.client.appid == "123"
        assert self.client.appsecret == "321"

    def test_robot_client(self):
        robot = ZgRoBot()
        assert robot.client.config == robot.config

    def test_robot_reuse_client(self):
        robot = ZgRoBot()
        client_1 = robot.client
        client_2 = robot.client
        assert client_1 is client_2

    def test_check_error(self):
        error_json = dict(error_code=0)
        assert error_json == check_error(error_json)

        error_json = dict(errcode=1, errmsg="test")
        with pytest.raises(ClientException) as err:
            check_error(error_json)
        assert err.value.args[0] == "1: test"

    @responses.activate
    @add_token_response
    def test_grant_token(self):
        # responses.add_callback(responses.GET, TOKEN_URL, callback=token_callback)
        self.client.grant_token()
        assert self.client.token == "ACCESS_TOKEN"

    @responses.activate
    @add_token_response
    def test_client_request(self):
        EMPTY_PARAMS_URL = "http://empty-params.zgrobot.com/"
        DATA_EXISTS_URL = "http://data-exists.zgrobot.com/"

        def empty_params_callback(request):
            params = urllib.parse.parse_qs(
                urllib.parse.urlparse(request.url).query
            )
            assert params["access_token"][0] == self.client.token
            return 200, JSON_HEADER, json.dumps({"test": "test"})

        def data_exists_url(request):
            assert json.loads(request.body.decode('utf-8')) == {"test": "test"}
            return 200, JSON_HEADER, json.dumps({"test": "test"})

        responses.add_callback(
            responses.POST, DATA_EXISTS_URL, callback=data_exists_url
        )
        responses.add_callback(
            responses.GET, EMPTY_PARAMS_URL, callback=empty_params_callback
        )
        responses.add_callback(
            responses.GET, TOKEN_URL, callback=token_callback
        )

        r = self.client.get(url=EMPTY_PARAMS_URL)
        assert r == {"test": "test"}

        r = self.client.post(url=DATA_EXISTS_URL, data={"test": "test"})
        assert r == {"test": "test"}


class TestClientBaseClassPost(TestClientBaseClass):
    @pytest.fixture(autouse=True)
    def mock_request(self, mocker):
        self.mocked_request = mocker.spy(self.client, 'request')

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_post_with_unittest(self):
        POST_FILE_URL = "http://post_file.zgrobot.com/"

        def empty_post_file_callback(request):
            return 200, JSON_HEADER, json.dumps({"test": "test"})

        responses.add_callback(
            responses.POST, POST_FILE_URL, callback=empty_post_file_callback
        )

        with open(GOD_PIC, 'rb') as f:
            self.client.post(url=POST_FILE_URL, files={"media": f})
            self.mocked_request.assert_any_call(
                method='post',
                url='http://post_file.zgrobot.com/',
                files=dict(media=(urllib.parse.quote(GOD_PIC), f))
            )

        # Test another attribute
        with open(GOD_PIC, 'rb') as f:
            self.client.post(url=POST_FILE_URL, files={"gugugu": f})
            self.mocked_request.assert_any_call(
                method='post',
                url='http://post_file.zgrobot.com/',
                files=dict(gugugu=(urllib.parse.quote(GOD_PIC), f))
            )

    @responses.activate
    @add_token_response
    def test_post_with_file_object_without_name(self):
        POST_FILE_URL = "http://post_file.zgrobot.com/"

        def empty_post_file_callback(request):
            return 200, JSON_HEADER, json.dumps({"test": "test"})

        responses.add_callback(
            responses.POST, POST_FILE_URL, callback=empty_post_file_callback
        )

        f = BytesIO(b'gugugu')
        self.client.post(url=POST_FILE_URL, files={"media": f})
        self.mocked_request.assert_any_call(
            method='post',
            url='http://post_file.zgrobot.com/',
            files=dict(media=f)
        )

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_post_with_integration_test(self):
        POST_FILE_URL = "http://post_file.zgrobot.com/"

        def post_file_callback(request):
            s = request.body.split(b"\r")[0][2:]
            p = list(
                multipart.MultipartParser(
                    BytesIO(multipart.tob(request.body)), s
                )
            )[0]
            assert "filename" in p.options
            return 200, JSON_HEADER, json.dumps({"test": "test"})

        responses.add_callback(
            responses.POST, POST_FILE_URL, callback=post_file_callback
        )

        with open(GOD_PIC, 'rb') as f:
            self.client.post(url=POST_FILE_URL, files={"media": f})


class TestClientMenuClass(BaseTestClass):
    CREATE_URL = "https://api.weixin.qq.com/cgi-bin/menu/create"
    GET_URL = "https://api.weixin.qq.com/cgi-bin/menu/get"
    DELETE_URL = "https://api.weixin.qq.com/cgi-bin/menu/delete"

    menu_data = {
        "button": [
            {
                "type": "click",
                "name": u"今日歌曲",
                "key": "V1001_TODAY_MUSIC"
            }, {
                "type": "click",
                "name": u"歌手简介",
                "key": "V1001_TODAY_SINGER"
            }, {
                "name": u"菜单",
                "sub_button": [
                    {
                        "type": "view",
                        "name": u"搜索",
                        "url": "http://www.soso.com/"
                    }, {
                        "type": "view",
                        "name": u"视频",
                        "url": "http://v.qq.com/"
                    }, {
                        "type": "click",
                        "name": u"赞一下我们",
                        "key": "V1001_GOOD"
                    }
                ]
            }
        ]
    }

    @staticmethod
    def create_menu_callback(request):
        def check_menu_data(item):
            keys = item.keys()
            assert "name" in keys
            if "sub_button" in keys:
                for button in item["sub_button"]:
                    check_menu_data(button)
                return
            assert "type" in keys
            if "type" == "click":
                assert "key" in keys
            elif "type" == "view":
                assert "url" in keys
            elif "type" == "media_id" or "type" == "view_limited":
                assert "media_id" in keys

        try:
            body = json.loads(request.body.decode("utf-8"))["button"]
        except KeyError:
            return 200, JSON_HEADER, json.dumps(
                {
                    "errcode": 1,
                    "errmsg": "error"
                }
            )
        try:
            for item in body:
                check_menu_data(item)
        except AssertionError:
            return 200, JSON_HEADER, json.dumps(
                {
                    "errcode": 1,
                    "errmsg": "error"
                }
            )
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_create_menu(self):
        responses.add_callback(
            responses.POST,
            self.CREATE_URL,
            callback=self.create_menu_callback
        )
        r = self.client.create_menu(self.menu_data)
        assert r == {"errcode": 0, "errmsg": "ok"}
        with pytest.raises(ClientException) as err:
            self.client.create_menu({"error": "error"})
        assert err.value.args[0] == "1: error"

    @responses.activate
    @add_token_response
    def test_get_menu(self):
        responses.add_callback(
            responses.GET, self.GET_URL, callback=self.callback_without_check
        )
        r = self.client.get_menu()
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_delete_menu(self):
        responses.add_callback(
            responses.GET,
            self.DELETE_URL,
            callback=self.callback_without_check
        )
        r = self.client.delete_menu()
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientGroupClass(BaseTestClass):
    CREATE_URL = "https://api.weixin.qq.com/cgi-bin/groups/create"
    GET_URL = "https://api.weixin.qq.com/cgi-bin/groups/get"
    GET_WITH_ID_URL = "https://api.weixin.qq.com/cgi-bin/groups/getid"
    UPDATE_URL = "https://api.weixin.qq.com/cgi-bin/groups/update"
    MOVE_URL = "https://api.weixin.qq.com/cgi-bin/groups/members/update"
    MOVE_USERS_URL = "https://api.weixin.qq.com/cgi-bin/groups/members/batchupdate"
    DELETE_URL = "https://api.weixin.qq.com/cgi-bin/groups/delete"

    @staticmethod
    def create_group_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "group" in body.keys()
        assert "name" in body["group"].keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def get_groups_with_id_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "openid" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def update_group_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "group" in body.keys()
        assert "id" in body["group"].keys()
        assert "name" in body["group"].keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def move_user_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "openid" in body.keys()
        assert "to_groupid" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def move_users_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "openid_list" in body.keys()
        assert "to_groupid" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def delete_group_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "group" in body.keys()
        assert "id" in body["group"].keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_create_group(self):
        responses.add_callback(
            responses.POST,
            self.CREATE_URL,
            callback=self.create_group_callback
        )
        r = self.client.create_group("test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_get_group(self):
        responses.add_callback(
            responses.GET, self.GET_URL, callback=self.callback_without_check
        )
        r = self.client.get_groups()
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_get_group_with_id(self):
        responses.add_callback(
            responses.POST,
            self.GET_WITH_ID_URL,
            callback=self.get_groups_with_id_callback
        )
        r = self.client.get_group_by_id("test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_update_group(self):
        responses.add_callback(
            responses.POST,
            self.UPDATE_URL,
            callback=self.update_group_callback
        )
        r = self.client.update_group("0", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_move_user(self):
        responses.add_callback(
            responses.POST, self.MOVE_URL, callback=self.move_user_callback
        )
        r = self.client.move_user("test", "0")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_move_users(self):
        responses.add_callback(
            responses.POST,
            self.MOVE_USERS_URL,
            callback=self.move_users_callback
        )
        r = self.client.move_users("test", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_delete_group(self):
        responses.add_callback(
            responses.POST,
            self.DELETE_URL,
            callback=self.delete_group_callback
        )
        r = self.client.delete_group("test")
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientRemarkClass(BaseTestClass):
    REMARK_URL = "https://api.weixin.qq.com/cgi-bin/user/info/updateremark"

    @staticmethod
    def remark_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "openid" in body.keys()
        assert "remark" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_client_remark(self):
        responses.add_callback(
            responses.POST, self.REMARK_URL, callback=self.remark_callback
        )
        r = self.client.remark_user("test", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientUserInfo(BaseTestClass):
    SINGLE_USER_URL = "https://api.weixin.qq.com/cgi-bin/user/info"
    MULTI_USER_URL = "https://api.weixin.qq.com/cgi-bin/user/info/batchget"

    @staticmethod
    def single_user_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        assert "openid" in params.keys()
        assert "lang" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def multi_user_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "user_list" in body.keys()
        for user in body["user_list"]:
            assert "openid" in user.keys()
            assert "lang" in user.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_single_user(self):
        responses.add_callback(
            responses.GET,
            self.SINGLE_USER_URL,
            callback=self.single_user_callback
        )
        r = self.client.get_user_info("test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_multi_user(self):
        responses.add_callback(
            responses.POST,
            self.MULTI_USER_URL,
            callback=self.multi_user_callback
        )
        r = self.client.get_users_info(["test1", "test2"])
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientGetFollowersClass(BaseTestClass):
    FOLLOWER_URL = "https://api.weixin.qq.com/cgi-bin/user/get"

    @staticmethod
    def get_followers_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        assert "next_openid" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_get_followers(self):
        responses.add_callback(
            responses.GET,
            self.FOLLOWER_URL,
            callback=self.get_followers_callback
        )
        r = self.client.get_followers("test")
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientCustomMenuClass(BaseTestClass):
    GET_URL = "https://api.weixin.qq.com/cgi-bin/get_current_selfmenu_info"
    CREATE_URL = "https://api.weixin.qq.com/cgi-bin/menu/addconditional"
    DELETE_URL = "https://api.weixin.qq.com/cgi-bin/menu/delconditional"
    MATCH_URL = "https://api.weixin.qq.com/cgi-bin/menu/trymatch"

    custom_data = {
        "menu_data": [
            {
                "type": "click",
                "name": u"今日歌曲",
                "key": "V1001_TODAY_MUSIC"
            }, {
                "name": u"菜单",
                "sub_button": [
                    {
                        "type": "view",
                        "name": u"搜索",
                        "url": "http://www.soso.com/"
                    }, {
                        "type": "view",
                        "name": u"视频",
                        "url": "http://v.qq.com/"
                    }, {
                        "type": "click",
                        "name": u"赞一下我们",
                        "key": "V1001_GOOD"
                    }
                ]
            }
        ],
        "matchrule": {
            "group_id": "2",
            "sex": "1",
            "country": u"中国",
            "province": u"广东",
            "city": u"广州",
            "client_platform_type": "2",
            "language": "zh_CN"
        }
    }

    @staticmethod
    def get_custom_menu_callback(request):
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def create_custom_menu_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "button" in body.keys()
        assert "matchrule" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def delete_custom_menu_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "menuid" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def match_custom_menu(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "user_id" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_get_custom_menu(self):
        responses.add_callback(
            responses.GET,
            self.GET_URL,
            callback=self.get_custom_menu_callback
        )
        r = self.client.get_custom_menu_config()
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_create_custom_menu(self):
        responses.add_callback(
            responses.POST,
            self.CREATE_URL,
            callback=self.create_custom_menu_callback
        )
        r = self.client.create_custom_menu(**self.custom_data)
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_delete_custom_menu(self):
        responses.add_callback(
            responses.POST,
            self.DELETE_URL,
            callback=self.delete_custom_menu_callback
        )
        r = self.client.delete_custom_menu("test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_march_custom_menu(self):
        responses.add_callback(
            responses.POST, self.MATCH_URL, callback=self.match_custom_menu
        )
        r = self.client.match_custom_menu("test")
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientResourceClass(BaseTestClass):
    UPLOAD_URL = "https://api.weixin.qq.com/cgi-bin/media/upload"
    DOWNLOAD_URL = "https://api.weixin.qq.com/cgi-bin/media/get"
    ADD_NEWS_URL = "https://api.weixin.qq.com/cgi-bin/material/add_news"
    UPLOAD_PICTURE_URL = "https://api.weixin.qq.com/cgi-bin/media/uploadimg"
    UPLOAD_P_URL = "https://api.weixin.qq.com/cgi-bin/material/add_material"
    DOWNLOAD_P_URL = "https://api.weixin.qq.com/cgi-bin/material/get_material"
    DELETE_P_URL = "https://api.weixin.qq.com/cgi-bin/material/del_material"
    UPDATE_NEWS_URL = "https://api.weixin.qq.com/cgi-bin/material/update_news"
    add_news_data = [
        {
            "title": "test_title",
            "thumb_media_id": "test",
            "author": "test",
            "digest": "test",
            "show_cover_pic": 1,
            "content": "test",
            "content_source_url": "test"
        }
    ]
    update_data = {
        "media_id": "test",
        "index": "test",
        "articles": {
            "title": "test",
            "thumb_media_id": "test",
            "author": "test",
            "digest": "test",
            "show_cover_pic": 1,
            "content": "test",
            "content_source_url": "test"
        }
    }

    @staticmethod
    def upload_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "type" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def download_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "media_id" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def add_news_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "articles" in body.keys()
        for article in body["articles"]:
            assert "title" in article.keys()
            assert "thumb_media_id" in article.keys()
            assert "author" in article.keys()
            assert "digest" in article.keys()
            assert "show_cover_pic" in article.keys()
            assert "content" in article.keys()
            assert "content_source_url" in article.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def upload_picture_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def upload_p_media_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        assert "type" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def download_p_media_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert "media_id" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def delete_p_media_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "media_id" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def update_news_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "media_id" in body.keys()
        assert "index" in body.keys()
        assert "articles" in body.keys()
        articles = body["articles"]
        assert "title" in articles.keys()
        assert "thumb_media_id" in articles.keys()
        assert "author" in articles.keys()
        assert "digest" in articles.keys()
        assert "show_cover_pic" in articles.keys()
        assert "content" in articles.keys()
        assert "content_source_url" in articles.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_upload_media(self):
        responses.add_callback(
            responses.POST, self.UPLOAD_URL, callback=self.upload_callback
        )
        with open(GOD_PIC, 'rb') as f:
            r = self.client.upload_media('image', f)
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_download_media(self):
        responses.add_callback(
            responses.GET, self.DOWNLOAD_URL, callback=self.download_callback
        )
        r = self.client.download_media("test")
        assert isinstance(r, requests.Response)

    @responses.activate
    @add_token_response
    def test_add_news(self):
        responses.add_callback(
            responses.POST, self.ADD_NEWS_URL, callback=self.add_news_callback
        )
        r = self.client.add_news(self.add_news_data)
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_upload_news_picture(self):
        responses.add_callback(
            responses.POST,
            self.UPLOAD_PICTURE_URL,
            callback=self.upload_picture_callback
        )
        with open(GOD_PIC, 'rb') as f:
            r = self.client.upload_news_picture(f)
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_upload_permanent_media(self):
        responses.add_callback(
            responses.POST,
            self.UPLOAD_P_URL,
            callback=self.upload_p_media_callback
        )
        with open(GOD_PIC, 'rb') as f:
            r = self.client.upload_permanent_media('image', f)
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_download_permanent_media(self):
        responses.add_callback(
            responses.POST,
            self.DOWNLOAD_P_URL,
            callback=self.download_p_media_callback
        )
        r = self.client.download_permanent_media("test")
        assert isinstance(r, requests.Response)

    @responses.activate
    @add_token_response
    def test_delete_permanent_media(self):
        responses.add_callback(
            responses.POST,
            self.DELETE_P_URL,
            callback=self.delete_p_media_callback
        )
        r = self.client.delete_permanent_media("test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_update_news(self):
        responses.add_callback(
            responses.POST,
            self.UPDATE_NEWS_URL,
            callback=self.update_news_callback
        )
        r = self.client.update_news(self.update_data)
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestUploadVideoClass(BaseTestClass):
    UPLOAD_VIDEO_URL = "https://api.weixin.qq.com/cgi-bin/material/add_material"

    @staticmethod
    def upload_video_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "type" in params.keys()
        assert params["type"][0] == "video"
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_upload_video(self):
        responses.add_callback(
            responses.POST,
            self.UPLOAD_VIDEO_URL,
            callback=self.upload_video_callback
        )
        with open(GOD_PIC, 'rb') as f:
            r = self.client.upload_permanent_video("test", "test", f)
        assert isinstance(r, requests.Response)


class TestMediaClass(BaseTestClass):
    GET_URL = "https://api.weixin.qq.com/cgi-bin/material/get_materialcount"
    GET_LIST_URL = "https://api.weixin.qq.com/cgi-bin/material/batchget_material"

    @staticmethod
    def get_media_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def get_media_list_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "type" in body.keys()
        assert "offset" in body.keys()
        assert "count" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_get_media(self):
        responses.add_callback(
            responses.GET, self.GET_URL, callback=self.get_media_callback
        )
        r = self.client.get_media_count()
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_get_media_list(self):
        responses.add_callback(
            responses.POST,
            self.GET_LIST_URL,
            callback=self.get_media_list_callback
        )
        r = self.client.get_media_list("test", "test", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestGetIpListClass(BaseTestClass):
    GET_URL = "https://api.weixin.qq.com/cgi-bin/getcallbackip"

    @staticmethod
    def get_ip_list_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_get_ip_list(self):
        responses.add_callback(
            responses.GET, self.GET_URL, callback=self.get_ip_list_callback
        )
        r = self.client.get_ip_list()
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestCustomService(BaseTestClass):
    ADD_URL = "https://api.weixin.qq.com/customservice/kfaccount/add"
    UPDATE_URL = "https://api.weixin.qq.com/customservice/kfaccount/update"
    DELETE_URL = "https://api.weixin.qq.com/customservice/kfaccount/del"
    UPLOAD_URL = "http://api.weixin.qq.com/customservice/kfaccount/uploadheadimg"
    GET_URL = "https://api.weixin.qq.com/cgi-bin/customservice/getkflist"
    GET_ONLINE_URL = "https://api.weixin.qq.com/cgi-bin/customservice/getonlinekflist"

    @staticmethod
    def add_update_delete_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "kf_account" in body.keys()
        assert "nickname" in body.keys()
        assert "password" in body.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def upload_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def get_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_add_custom_service_account(self):
        responses.add_callback(
            responses.POST,
            self.ADD_URL,
            callback=self.add_update_delete_callback
        )
        r = self.client.add_custom_service_account("test", "test", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_update_custom_service_account(self):
        responses.add_callback(
            responses.POST,
            self.UPDATE_URL,
            callback=self.add_update_delete_callback
        )
        r = self.client.update_custom_service_account("test", "test", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_delete_custom_service_account(self):
        responses.add_callback(
            responses.POST,
            self.DELETE_URL,
            callback=self.add_update_delete_callback
        )
        r = self.client.delete_custom_service_account("test", "test", "test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    @create_pic_file
    def test_upload_custom_service_account_avatar(self):
        responses.add_callback(
            responses.POST, self.UPLOAD_URL, callback=self.upload_callback
        )
        with open(GOD_PIC, 'rb') as f:
            r = self.client.upload_custom_service_account_avatar("image", f)
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_get_custom_service_account_list(self):
        responses.add_callback(
            responses.GET, self.GET_URL, callback=self.get_callback
        )
        r = self.client.get_custom_service_account_list()
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_get_online_custom_service_account_list(self):
        responses.add_callback(
            responses.GET, self.GET_ONLINE_URL, callback=self.get_callback
        )
        r = self.client.get_online_custom_service_account_list()
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestQrcodeClass(BaseTestClass):
    CREATE_URL = "https://api.weixin.qq.com/cgi-bin/qrcode/create"
    SHOW_URL = "https://mp.weixin.qq.com/cgi-bin/showqrcode"

    @staticmethod
    def create_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @staticmethod
    def show_callback(request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "ticket" in params.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_create_qrcode(self):
        responses.add_callback(
            responses.POST, self.CREATE_URL, callback=self.create_callback
        )
        r = self.client.create_qrcode("test")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @responses.activate
    @add_token_response
    def test_show_qrcode(self):
        responses.add_callback(
            responses.GET, self.SHOW_URL, callback=self.show_callback
        )
        r = self.client.show_qrcode("test")
        assert isinstance(r, requests.Response)


class TestSendArticleMessagesClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def article_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "news"
        assert "news" in body.keys()
        for article in body["news"]["articles"]:
            assert "title" in article.keys()
            assert "description" in article.keys()
            assert "url" in article.keys()
            assert "picurl" in article.keys()
        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_article_messages(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.article_callback
        )

        from zgrobot.replies import Article
        articles = []
        for _ in range(0, 8):
            articles.append(
                Article(
                    *[
                        "test_title",
                        "test_description",
                        "test_img",
                        "test_url",
                    ]
                )
            )

        r = self.client.send_article_message("test_id", articles)
        assert r == {"errcode": 0, "errmsg": "ok"}

        articles = []
        for _ in range(0, 8):
            articles.append(
                {
                    "title": "test_title",
                    "description": "test_description",
                    "url": "test_url",
                    "picurl": "test_pic_url"
                }
            )

        r = self.client.send_article_message("test_id", articles)
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestSendTextMessageClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def text_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "text"
        assert "text" in body.keys()
        assert "content" in body["text"].keys()
        assert "customservice" not in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_text_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.text_callback
        )

        r = self.client.send_text_message("test_id", "test_message")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @staticmethod
    def text_with_kf_account_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "text"
        assert "text" in body.keys()
        assert "content" in body["text"].keys()
        assert "customservice" in body.keys()
        assert "kf_account" in body["customservice"].keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_text_message_with_kf_account(self):
        responses.add_callback(
            responses.POST,
            self.URL,
            callback=self.text_with_kf_account_callback
        )
        r = self.client.send_text_message("test_id", "test_message", "233@233")
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestSendImageMessageClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def image_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "image"
        assert "image" in body.keys()
        assert "media_id" in body["image"].keys()
        assert "customservice" not in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_image_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.image_callback
        )

        r = self.client.send_image_message("test_id", "test_media_id")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @staticmethod
    def image_with_kf_account_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "image"
        assert "image" in body.keys()
        assert "media_id" in body["image"].keys()
        assert "customservice" in body.keys()
        assert "kf_account" in body["customservice"].keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_image_message_with_kf_account(self):
        responses.add_callback(
            responses.POST,
            self.URL,
            callback=self.image_with_kf_account_callback
        )

        r = self.client.send_image_message(
            "test_id", "test_media_id", "233@233"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestSendVoiceMessageClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def voice_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "voice"
        assert "voice" in body.keys()
        assert "media_id" in body["voice"].keys()
        assert "customservice" not in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_voice_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.voice_callback
        )

        r = self.client.send_voice_message("test_id", "test_media_id")
        assert r == {"errcode": 0, "errmsg": "ok"}

    @staticmethod
    def voice_with_kf_account_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "voice"
        assert "voice" in body.keys()
        assert "media_id" in body["voice"].keys()
        assert "customservice" in body.keys()
        assert "kf_account" in body["customservice"].keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_voice_message_with_kf_account(self):
        responses.add_callback(
            responses.POST,
            self.URL,
            callback=self.voice_with_kf_account_callback
        )

        r = self.client.send_voice_message(
            "test_id", "test_media_id", "233@233"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestMusicMessageClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def music_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "music"
        assert "music" in body.keys()
        assert "musicurl" in body["music"].keys()
        assert "hqmusicurl" in body["music"].keys()
        assert "thumb_media_id" in body["music"].keys()
        assert "customservice" not in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_music_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.music_callback
        )

        r = self.client.send_music_message(
            user_id="test_id",
            url="test_url",
            hq_url="test_hq_url",
            thumb_media_id="test_media_id",
            title="test_title",
            description="test_description"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}

    @staticmethod
    def music_with_kf_account_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "music"
        assert "music" in body.keys()
        assert "musicurl" in body["music"].keys()
        assert "hqmusicurl" in body["music"].keys()
        assert "thumb_media_id" in body["music"].keys()
        assert "customservice" in body.keys()
        assert "kf_account" in body["customservice"].keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_music_message_with_kf_account(self):
        responses.add_callback(
            responses.POST,
            self.URL,
            callback=self.music_with_kf_account_callback
        )

        r = self.client.send_music_message(
            user_id="test_id",
            url="test_url",
            hq_url="test_hq_url",
            thumb_media_id="test_media_id",
            title="test_title",
            description="test_description",
            kf_account="233@233"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestVideoMessageClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def video_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "video"
        assert "video" in body.keys()
        assert "media_id" in body["video"].keys()
        assert "customservice" not in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_video_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.video_callback
        )
        r = self.client.send_video_message(
            user_id="test_id",
            media_id="test_media_id",
            title="test_title",
            description="test_description",
        )
        assert r == {"errcode": 0, "errmsg": "ok"}

    @staticmethod
    def video_with_kf_account_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "video"
        assert "video" in body.keys()
        assert "media_id" in body["video"].keys()
        assert "customservice" in body.keys()
        assert "kf_account" in body["customservice"].keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_video_message_with_kf_account(self):
        responses.add_callback(
            responses.POST,
            self.URL,
            callback=self.video_with_kf_account_callback
        )
        r = self.client.send_video_message(
            user_id="test_id",
            media_id="test_media_id",
            title="test_title",
            description="test_description",
            kf_account="233@233"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestNewsMessageClass(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def news_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "mpnews"
        assert "mpnews" in body.keys()
        assert "media_id" in body["mpnews"].keys()
        assert "customservice" not in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_news_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.news_callback
        )

        r = self.client.send_news_message(
            user_id="test_id", media_id="test_media_id"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}

    @staticmethod
    def news_with_kf_account_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "mpnews"
        assert "mpnews" in body.keys()
        assert "media_id" in body["mpnews"].keys()
        assert "customservice" in body.keys()
        assert "kf_account" in body["customservice"].keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_news_message_with_kf_account(self):
        responses.add_callback(
            responses.POST,
            self.URL,
            callback=self.news_with_kf_account_callback
        )

        r = self.client.send_news_message(
            user_id="test_id", media_id="test_media_id", kf_account="233@233"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestTemplateMessage(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/template/send"

    @staticmethod
    def template_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "template_id" in body.keys()
        assert "url" in body.keys()
        assert "data" in body.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_template_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.template_callback
        )

        r = self.client.send_template_message(
            user_id="test_id",
            template_id="test_template_id",
            data="test_data",
            url="test_url"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestMiniprogrampageMessage(BaseTestClass):
    URL = "https://api.weixin.qq.com/cgi-bin/message/custom/send"

    @staticmethod
    def miniprogrampage_callback(request):
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body.keys()
        assert "msgtype" in body.keys()
        assert body["msgtype"] == "miniprogrampage"
        assert "miniprogrampage" in body.keys()
        miniprogrampage = body["miniprogrampage"]
        assert "title" in miniprogrampage.keys()
        assert "appid" in miniprogrampage.keys()
        assert "pagepath" in miniprogrampage.keys()
        assert "thumb_media_id" in miniprogrampage.keys()

        return 200, JSON_HEADER, json.dumps({"errcode": 0, "errmsg": "ok"})

    @responses.activate
    @add_token_response
    def test_send_miniprogrampage_message(self):
        responses.add_callback(
            responses.POST, self.URL, callback=self.miniprogrampage_callback
        )

        r = self.client.send_miniprogrampage_message(
            user_id="test_id",
            title="test_title",
            appid="test_appid",
            pagepath="test_pagepath",
            thumb_media_id="test_id"
        )
        assert r == {"errcode": 0, "errmsg": "ok"}


class TestClientTagManageClass(BaseTestClass):
    CREATE_TAG_URL = "https://api.weixin.qq.com/cgi-bin/tags/create"
    GET_TAGS_URL = "https://api.weixin.qq.com/cgi-bin/tags/get"
    UPDATE_TAG_URL = "https://api.weixin.qq.com/cgi-bin/tags/update"
    DELETE_TAG_URL = "https://api.weixin.qq.com/cgi-bin/tags/delete"
    GET_USERS_BY_TAG_URL = "https://api.weixin.qq.com/cgi-bin/user/tag/get"

    create_tag_name = u"喵喵喵"
    create_tag_id = 100

    update_tag_name = "nyanya"
    update_tag_id = 100

    get_tags_response = {
        'tags': [
            {
                'id': 2,
                'name': u'星标组',
                'count': 0
            }, {
                'id': 100,
                'name': update_tag_name,
                'count': 0
            }
        ]
    }

    get_users_by_tag_id = 100
    get_users_by_tag_response = {
        'count': 1,
        'data': {
            'openid': ['testopenid']
        },
        'next_openid': 'testopenid'
    }

    delete_tag_id = 100

    def create_tag_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {"tag": {"name": self.create_tag_name}}
        return 200, JSON_HEADER, json.dumps(
            {"tag": {
                "id": self.create_tag_id,
                "name": self.create_tag_name
            }}
        )

    def update_tag_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {
            "tag": {
                "id": self.update_tag_id,
                "name": self.update_tag_name
            }
        }
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def get_tags_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps(self.get_tags_response)

    def get_users_by_tag_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {"tagid": self.get_users_by_tag_id, "next_openid": ""}
        return 200, JSON_HEADER, json.dumps(self.get_users_by_tag_response)

    def delete_tag_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {
            "tag": {
                "id": self.delete_tag_id,
            }
        }
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    @responses.activate
    @add_token_response
    def test_create_tag(self):
        responses.add_callback(
            responses.POST,
            self.CREATE_TAG_URL,
            callback=self.create_tag_callback
        )
        r = self.client.create_tag(self.create_tag_name)
        assert r == {
            "tag": {
                "id": self.create_tag_id,
                "name": self.create_tag_name
            }
        }

    @responses.activate
    @add_token_response
    def test_update_tag(self):
        responses.add_callback(
            responses.POST,
            self.UPDATE_TAG_URL,
            callback=self.update_tag_callback
        )
        r = self.client.update_tag(self.update_tag_id, self.update_tag_name)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_get_tags(self):
        responses.add_callback(
            responses.GET, self.GET_TAGS_URL, callback=self.get_tags_callback
        )
        r = self.client.get_tags()
        assert r == self.get_tags_response

    @responses.activate
    @add_token_response
    def test_get_users_by_tag(self):
        responses.add_callback(
            responses.POST,
            self.GET_USERS_BY_TAG_URL,
            callback=self.get_users_by_tag_callback
        )
        r = self.client.get_users_by_tag(self.get_users_by_tag_id)
        assert r == self.get_users_by_tag_response

    @responses.activate
    @add_token_response
    def test_delete_tag(self):
        responses.add_callback(
            responses.POST,
            self.DELETE_TAG_URL,
            callback=self.delete_tag_callback
        )
        r = self.client.delete_tag(self.delete_tag_id)
        assert r == {'errcode': 0, 'errmsg': 'ok'}


class TestClientMembersTagClass(BaseTestClass):
    TAG_USER_URL = "https://api.weixin.qq.com/cgi-bin/tags/members/batchtagging"
    UNTAG_USER_URL = "https://api.weixin.qq.com/cgi-bin/tags/members/batchuntagging"
    GET_TAGS_BY_USER_URL = "https://api.weixin.qq.com/cgi-bin/tags/getidlist"

    tag_id = 100
    user_a_open_id = 'a'
    user_b_open_id = 'b'
    user_c_open_id = 'c'
    users_list = [user_a_open_id, user_b_open_id, user_c_open_id]

    get_tags_by_user_response = {'tagid_list': [tag_id]}

    def tag_user_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {"openid_list": self.users_list, "tagid": self.tag_id}
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def untag_user_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {"openid_list": self.users_list, "tagid": self.tag_id}
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def get_tags_by_user_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {
            "openid": self.user_a_open_id,
        }
        return 200, JSON_HEADER, json.dumps(self.get_tags_by_user_response)

    @responses.activate
    @add_token_response
    def test_tag_users(self):
        responses.add_callback(
            responses.POST, self.TAG_USER_URL, callback=self.tag_user_callback
        )
        r = self.client.tag_users(self.tag_id, self.users_list)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_untag_users(self):
        responses.add_callback(
            responses.POST,
            self.UNTAG_USER_URL,
            callback=self.untag_user_callback
        )
        r = self.client.untag_users(self.tag_id, self.users_list)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_get_tags_by_user(self):
        responses.add_callback(
            responses.POST,
            self.GET_TAGS_BY_USER_URL,
            callback=self.get_tags_by_user_callback
        )
        r = self.client.get_tags_by_user(self.user_a_open_id)
        assert r == self.get_tags_by_user_response


class TestClientMass(BaseTestClass):
    UP_NEWS_URL = 'https://api.weixin.qq.com/cgi-bin/media/uploadnews'
    SEND_ALL_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/sendall'
    SEND_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/send'
    DELETE_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/delete'
    PREVIEW_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/preview'
    GET_STATUS_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/get'
    GET_SPEED_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/speed/get'
    SET_SPEED_URL = 'https://api.weixin.qq.com/cgi-bin/message/mass/speed/set'

    articles = [
        {
            "thumb_media_id": "qI6_Ze_6PtV7svjolgs-rN6stStuHIjs9_DidOHaj0Q-mwvBelOXCFZiq2OsIU-p",
            "author": "xxx",
            "title": "Happy Day",
            "content_source_url": "www.qq.com",
            "content": "content",
            "digest": "digest",
            "show_cover_pic": 1,
            "need_open_comment": 1,
            "only_fans_can_comment": 1
        }, {
            "thumb_media_id": "qI6_Ze_6PtV7svjolgs-rN6stStuHIjs9_DidOHaj0Q-mwvBelOXCFZiq2OsIU-p",
            "author": "xxx",
            "title": "Happy Day",
            "content_source_url": "www.qq.com",
            "content": "content",
            "digest": "digest",
            "show_cover_pic": 0,
            "need_open_comment": 1,
            "only_fans_can_comment": 1
        }
    ]

    def up_news_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert body == {"articles": self.articles}
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def send_all_openid_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert 'touser' in body
        count = len(body['touser'])
        assert count >= 2
        assert count <= 10000
        assert 'msgtype' in body
        types = {
            'mpnews': 'media_id',
            'mpvideo': 'media_id',
            'music': 'media_id',
            'image': 'media_id',
            'video': 'media_id',
            'wxcard': 'card_id',
            'text': 'content',
        }
        assert types[body['msgtype']] in body[body['msgtype']]
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def send_all_tagid_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert 'filter' in body
        assert 'msgtype' in body
        types = {
            'mpnews': 'media_id',
            'mpvideo': 'media_id',
            'music': 'media_id',
            'image': 'media_id',
            'video': 'media_id',
            'wxcard': 'card_id',
            'text': 'content',
        }
        assert types[body['msgtype']] in body[body['msgtype']]
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def delete_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert "msg_id" in body
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def preview_openid_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert "touser" in body
        types = {
            'mpnews': 'media_id',
            'mpvideo': 'media_id',
            'music': 'media_id',
            'image': 'media_id',
            'video': 'media_id',
            'wxcard': 'card_id',
            'text': 'content',
        }
        assert types[body['msgtype']] in body[body['msgtype']]
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def preview_wxname_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert "towxname" in body
        types = {
            'mpnews': 'media_id',
            'mpvideo': 'media_id',
            'music': 'media_id',
            'image': 'media_id',
            'video': 'media_id',
            'wxcard': 'card_id',
            'text': 'content',
            'voice': 'media_id'
        }
        assert types[body['msgtype']] in body[body['msgtype']]
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    def get_status_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert "msg_id" in body
        return 200, JSON_HEADER, json.dumps(
            {
                "msg_id": body['msg_id'],
                "msg_status": "SEND_SUCCESS"
            }
        )

    def get_news_speed_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        return 200, JSON_HEADER, json.dumps({"speed": 3, "realspeed": 15})

    def set_news_speed_callback(self, request):
        params = urllib.parse.parse_qs(
            urllib.parse.urlparse(request.url).query
        )
        assert "access_token" in params.keys()
        body = json.loads(request.body.decode("utf-8"))
        assert "speed" in body
        return 200, JSON_HEADER, json.dumps({'errcode': 0, 'errmsg': 'ok'})

    @responses.activate
    @add_token_response
    def test_upload_news(self):
        responses.add_callback(
            responses.POST, self.UP_NEWS_URL, callback=self.up_news_callback
        )
        r = self.client.upload_news(self.articles)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_send_by_openid(self):
        responses.add_callback(
            responses.POST,
            self.SEND_URL,
            callback=self.send_all_openid_callback
        )
        r = self.client.send_mass_msg('mpnews', 'hgbkjnlkmlkn', ['1', '2'], 0)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_send_by_tagid(self):
        responses.add_callback(
            responses.POST,
            self.SEND_ALL_URL,
            callback=self.send_all_tagid_callback
        )
        r = self.client.send_mass_msg('mpnews', 'hgbkjnlkmlkn', 2, 0)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_send_all(self):
        responses.add_callback(
            responses.POST,
            self.SEND_ALL_URL,
            callback=self.send_all_tagid_callback
        )
        r = self.client.send_mass_msg('mpnews', 'hgbkjnlkmlkn', 0, 0)
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_preview_openid(self):
        responses.add_callback(
            responses.POST,
            self.PREVIEW_URL,
            callback=self.preview_openid_callback
        )
        r = self.client.send_mass_preview_to_user(
            type, 'hgbkjnlkmlkn', 'asdasdasd', 'openid'
        )
        assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_preview_wxname(self):
        responses.add_callback(
            responses.POST,
            self.PREVIEW_URL,
            callback=self.preview_wxname_callback
        )
        type_list = [
            'mpnews', 'text', 'voice', 'music', 'image', 'video', 'wxcard'
        ]
        for type in type_list:
            r = self.client.send_mass_preview_to_user(
                type, 'hgbkjnlkmlkn', 'asdasdasd', 'wxname'
            )
            assert r == {'errcode': 0, 'errmsg': 'ok'}

    @responses.activate
    @add_token_response
    def test_get_status(self):
        responses.add_callback(
            responses.POST,
            self.GET_STATUS_URL,
            callback=self.get_status_callback
        )
        r = self.client.get_mass_msg_status('mpnews')
        assert 'msg_id' in r
        assert 'msg_status' in r

    @responses.activate
    @add_token_response
    def test_get_speed(self):
        responses.add_callback(
            responses.POST,
            self.GET_SPEED_URL,
            callback=self.get_news_speed_callback
        )
        r = self.client.get_mass_msg_speed()
        assert 'speed' in r
        assert 'realspeed' in r

    @responses.activate
    @add_token_response
    def test_set_speed(self):
        responses.add_callback(
            responses.POST,
            self.SET_SPEED_URL,
            callback=self.set_news_speed_callback
        )
        r = self.client.set_mass_msg_speed(1)
        assert r == {'errcode': 0, 'errmsg': 'ok'}



# -*- coding: utf-8 -*-

from zgrobot.parser import parse_user_msg


def test_none_message():
    assert not parse_user_msg("")


def test_text_message():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>1348831860</CreateTime>
        <MsgType><![CDATA[text]]></MsgType>
        <Content><![CDATA[this is a test]]></Content>
        <MsgId>1234567890123456</MsgId>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 1348831860
    assert message.type == "text"
    assert message.content == "this is a test"
    assert message.message_id == 1234567890123456


def test_image_message():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>1348831860</CreateTime>
        <MsgType><![CDATA[image]]></MsgType>
        <PicUrl><![CDATA[this is a url]]></PicUrl>
        <MsgId>1234567890123456</MsgId>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 1348831860
    assert message.type == "image"
    assert message.img == "this is a url"
    assert message.message_id == 1234567890123456


def test_location_message():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>1351776360</CreateTime>
        <MsgType><![CDATA[location]]></MsgType>
        <Location_X>23.134521</Location_X>
        <Location_Y>113.358803</Location_Y>
        <Scale>20</Scale>
        <Label><![CDATA[Location]]></Label>
        <MsgId>1234567890123456</MsgId>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 1351776360
    assert message.type == "location"
    assert message.location == (23.134521, 113.358803)
    assert message.scale == 20
    assert message.label == "Location"
    assert message.message_id == 1234567890123456


def test_link_message():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>1351776360</CreateTime>
        <MsgType><![CDATA[link]]></MsgType>
        <Title><![CDATA[ZgRoBot]]></Title>
        <Description><![CDATA[Link to ZgRoBot]]></Description>
        <Url><![CDATA[https://github.com/whtsky/ZgRoBot]]></Url>
        <MsgId>1234567890123456</MsgId>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 1351776360
    assert message.type == "link"
    assert message.title == "ZgRoBot"
    assert message.description == "Link to ZgRoBot"
    assert message.url == "https://github.com/whtsky/ZgRoBot"
    assert message.message_id == 1234567890123456


def test_voice_message():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>1357290913</CreateTime>
        <MsgType><![CDATA[voice]]></MsgType>
        <MediaId><![CDATA[media_id]]></MediaId>
        <Format><![CDATA[Format]]></Format>
        <Recognition><![CDATA[Meow~]]></Recognition>
        <MsgId>1234567890123456</MsgId>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 1357290913
    assert message.type == "voice"
    assert message.media_id == "media_id"
    assert message.format == "Format"
    assert message.recognition == "Meow~"
    assert message.message_id == 1234567890123456


def test_unknown_message():
    xml = """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>1351776360</CreateTime>
        <MsgType><![CDATA[unknown]]></MsgType>
        <Title><![CDATA[ZgRoBot]]></Title>
        <Description><![CDATA[Link to ZgRoBot]]></Description>
        <Url><![CDATA[https://github.com/whtsky/ZgRoBot]]></Url>
        <MsgId>1234567890123456</MsgId>
    </xml>
    """
    message = parse_user_msg(xml)
    assert message.raw == xml
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 1351776360


def test_subscribe_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[subscribe]]></Event>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "subscribe_event"

    message = parse_user_msg(
        """
    <xml><ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[subscribe]]></Event>
        <EventKey><![CDATA[qrscene_123123]]></EventKey>
        <Ticket><![CDATA[TICKET]]></Ticket>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "subscribe_event"
    assert message.key == "qrscene_123123"
    assert message.ticket == "TICKET"


def test_unsubscribe_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[unsubscribe]]></Event>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 123456789
    assert message.type == "unsubscribe_event"


def test_scan_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[SCAN]]></Event>
        <EventKey><![CDATA[SCENE_VALUE]]></EventKey>
        <Ticket><![CDATA[TICKET]]></Ticket>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "scan_event"
    assert message.key == "SCENE_VALUE"
    assert message.ticket == "TICKET"


def test_scan_push_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[scancode_push]]></Event>
        <EventKey><![CDATA[EVENTKEY]]></EventKey>
        <ScanCodeInfo>
            <ScanType><![CDATA[qrcode]]></ScanType>
            <ScanResult><![CDATA[www.qq.com]]></ScanResult>
        </ScanCodeInfo>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "scancode_push_event"
    assert message.key == "EVENTKEY"
    assert message.scan_type == "qrcode"
    assert message.scan_result == "www.qq.com"


def test_scan_waitmsg_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[scancode_waitmsg]]></Event>
        <EventKey><![CDATA[EVENTKEY]]></EventKey>
        <ScanCodeInfo>
            <ScanType><![CDATA[qrcode]]></ScanType>
            <ScanResult><![CDATA[www.qq.com]]></ScanResult>
        </ScanCodeInfo>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "scancode_waitmsg_event"
    assert message.key == "EVENTKEY"
    assert message.scan_type == "qrcode"
    assert message.scan_result == "www.qq.com"


def test_pic_sysphoto_event():
    # count is 1
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_e136c6e50636]]></ToUserName>
        <FromUserName><![CDATA[oMgHVjngRipVsoxg6TuX3vz6glDg]]></FromUserName>
        <CreateTime>1408090651</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[pic_sysphoto]]></Event>
        <EventKey><![CDATA[6]]></EventKey>
        <SendPicsInfo>
            <Count>1</Count>
            <PicList>
                <item>
                    <PicMd5Sum><![CDATA[1b5f7c23b5bf75682a53e7b6d163e185]]></PicMd5Sum>
                </item>
            </PicList>
        </SendPicsInfo>
    </xml>
    """
    )
    assert message.target == "gh_e136c6e50636"
    assert message.source == "oMgHVjngRipVsoxg6TuX3vz6glDg"
    assert message.time == 1408090651
    assert message.type == "pic_sysphoto_event"
    assert message.key == "6"
    assert message.count == 1
    assert message.pic_list == [
        {
            'pic_md5_sum': '1b5f7c23b5bf75682a53e7b6d163e185'
        }
    ]

    # count is more than 1
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[gh_e136c6e50636]]></ToUserName>
            <FromUserName><![CDATA[oMgHVjngRipVsoxg6TuX3vz6glDg]]></FromUserName>
            <CreateTime>1408090651</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[pic_sysphoto]]></Event>
            <EventKey><![CDATA[6]]></EventKey>
            <SendPicsInfo>
                <Count>2</Count>
                <PicList>
                    <item>
                        <PicMd5Sum><![CDATA[1b5f7c23b5bf75682a53e7b6d163e185]]></PicMd5Sum>
                    </item>
                    <item>
                        <PicMd5Sum><![CDATA[233]]></PicMd5Sum>
                    </item>
                </PicList>
            </SendPicsInfo>
        </xml>
        """
    )
    assert message.target == "gh_e136c6e50636"
    assert message.source == "oMgHVjngRipVsoxg6TuX3vz6glDg"
    assert message.time == 1408090651
    assert message.type == "pic_sysphoto_event"
    assert message.key == "6"
    assert message.count == 2
    assert message.pic_list == [
        {
            'pic_md5_sum': '1b5f7c23b5bf75682a53e7b6d163e185'
        }, {
            'pic_md5_sum': '233'
        }
    ]


def test_pic_photo_or_album_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_e136c6e50636]]></ToUserName>
        <FromUserName><![CDATA[oMgHVjngRipVsoxg6TuX3vz6glDg]]></FromUserName>
        <CreateTime>1408090816</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[pic_photo_or_album]]></Event>
        <EventKey><![CDATA[6]]></EventKey>
        <SendPicsInfo>
            <Count>1</Count>
            <PicList>
                <item>
                    <PicMd5Sum><![CDATA[5a75aaca956d97be686719218f275c6b]]></PicMd5Sum>
                </item>
            </PicList>
        </SendPicsInfo>
    </xml>
    """
    )
    assert message.target == "gh_e136c6e50636"
    assert message.source == "oMgHVjngRipVsoxg6TuX3vz6glDg"
    assert message.time == 1408090816
    assert message.type == "pic_photo_or_album_event"
    assert message.key == "6"
    assert message.count == 1
    assert message.pic_list == [
        {
            'pic_md5_sum': '5a75aaca956d97be686719218f275c6b'
        }
    ]


def test_card_pass_check_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[card_pass_check]]></Event>
        <CardId><![CDATA[cardid]]></CardId>
        <RefuseReason><![CDATA[非法代制]]></RefuseReason>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "card_pass_check_event"
    assert message.card_id == "cardid"
    assert message.refuse_reason == u"非法代制"


def test_card_not_pass_check_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[toUser]]></ToUserName>
            <FromUserName><![CDATA[FromUser]]></FromUserName>
            <CreateTime>123456789</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[card_not_pass_check]]></Event>
            <CardId><![CDATA[cardid]]></CardId>
            <RefuseReason><![CDATA[非法代制]]></RefuseReason>
        </xml>
        """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "card_not_pass_check_event"
    assert message.card_id == "cardid"
    assert message.refuse_reason == u"非法代制"


def test_user_get_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName> <![CDATA[gh_fc0a06a20993]]> </ToUserName>
            <FromUserName> <![CDATA[oZI8Fj040-be6rlDohc6gkoPOQTQ]]> </FromUserName>
            <CreateTime>1472551036</CreateTime>
            <MsgType> <![CDATA[event]]> </MsgType>
            <Event> <![CDATA[user_get_card]]> </Event>
            <CardId> <![CDATA[pZI8Fjwsy5fVPRBeD78J4RmqVvBc]]> </CardId>
            <IsGiveByFriend>0</IsGiveByFriend>
            <UserCardCode> <![CDATA[226009850808]]> </UserCardCode>
            <FriendUserName> <![CDATA[]]> </FriendUserName>
            <OuterId>0</OuterId>
            <OldUserCardCode> <![CDATA[]]> </OldUserCardCode>
            <OuterStr> <![CDATA[12b]]> </OuterStr>
            <IsRestoreMemberCard>0</IsRestoreMemberCard>
            <IsRecommendByFriend>0</IsRecommendByFriend>
        </xml>
    """
    )
    assert message.target == "gh_fc0a06a20993"
    assert message.source == "oZI8Fj040-be6rlDohc6gkoPOQTQ"
    assert message.time == 1472551036
    assert message.type == "user_get_card_event"
    assert message.card_id == "pZI8Fjwsy5fVPRBeD78J4RmqVvBc"
    assert message.is_give_by_friend == 0
    assert message.user_card_code == "226009850808"
    assert message.friend_user_name is None
    assert message.outer_id == 0
    assert message.old_user_card_code is None
    assert message.outer_str == '12b'
    assert message.is_restore_member_card == 0
    assert message.is_recommend_by_friend == 0


def test_user_gifting_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[gh_3fcea188bf78]]></ToUserName>
            <FromUserName><![CDATA[obLatjjwDolFjRRd3doGIdwNqRXw]]></FromUserName>
            <CreateTime>1474181868</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[user_gifting_card]]></Event>
            <CardId><![CDATA[pbLatjhU-3pik3d4PsbVzvBxZvJc]]></CardId>
            <UserCardCode><![CDATA[297466945104]]></UserCardCode>
            <IsReturnBack>0</IsReturnBack>
            <FriendUserName><![CDATA[obLatjlNerkb62HtSdQUx66C4NTU]]></FriendUserName>
            <IsChatRoom>0</IsChatRoom>
        </xml>
    """
    )
    assert message.target == "gh_3fcea188bf78"
    assert message.source == "obLatjjwDolFjRRd3doGIdwNqRXw"
    assert message.time == 1474181868
    assert message.type == "user_gifting_card_event"
    assert message.card_id == "pbLatjhU-3pik3d4PsbVzvBxZvJc"
    assert message.user_card_code == "297466945104"
    assert message.is_return_back == 0
    assert message.friend_user_name == "obLatjlNerkb62HtSdQUx66C4NTU"
    assert message.is_chat_room == 0


def test_user_del_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[toUser]]></ToUserName>
            <FromUserName><![CDATA[FromUser]]></FromUserName>
            <CreateTime>123456789</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[user_del_card]]></Event>
            <CardId><![CDATA[cardid]]></CardId>
            <UserCardCode><![CDATA[12312312]]></UserCardCode>
        </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "user_del_card_event"
    assert message.card_id == "cardid"
    assert message.user_card_code == "12312312"


def test_user_consume_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName> <![CDATA[gh_fc0a06a20993]]> </ToUserName>
            <FromUserName> <![CDATA[oZI8Fj040-be6rlDohc6gkoPOQTQ]]> </FromUserName>
            <CreateTime>1472549042</CreateTime>
            <MsgType> <![CDATA[event]]> </MsgType>
            <Event> <![CDATA[user_consume_card]]> </Event>
            <CardId> <![CDATA[pZI8Fj8y-E8hpvho2d1ZvpGwQBvA]]> </CardId>
            <UserCardCode> <![CDATA[452998530302]]> </UserCardCode>
            <ConsumeSource> <![CDATA[FROM_API]]> </ConsumeSource>
            <LocationName> <![CDATA[]]> </LocationName>
            <StaffOpenId> <![CDATA[oZ********nJ3bPJu_Rtjkw4c]]> </StaffOpenId>
            <VerifyCode> <![CDATA[]]> </VerifyCode>
            <RemarkAmount> <![CDATA[]]> </RemarkAmount>
            <OuterStr> <![CDATA[xxxxx]]> </OuterStr>
        </xml>
    """
    )
    assert message.target == "gh_fc0a06a20993"
    assert message.source == "oZI8Fj040-be6rlDohc6gkoPOQTQ"
    assert message.time == 1472549042
    assert message.type == "user_consume_card_event"
    assert message.card_id == "pZI8Fj8y-E8hpvho2d1ZvpGwQBvA"
    assert message.user_card_code == "452998530302"
    assert message.consume_source == "FROM_API"
    assert message.location_name is None
    assert message.staff_open_id == "oZ********nJ3bPJu_Rtjkw4c"
    assert message.verify_code is None
    assert message.remark_amount is None
    assert message.outer_str == "xxxxx"


def test_user_pay_from_pay_cell_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[gh_e2243xxxxxxx]]></ToUserName>
            <FromUserName><![CDATA[oo2VNuOUuZGMxxxxxxxx]]></FromUserName>
            <CreateTime>1442390947</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[user_pay_from_pay_cell]]></Event>
            <CardId><![CDATA[po2VNuCuRo-8sxxxxxxxxxxx]]></CardId>
            <UserCardCode><![CDATA[38050000000]]></UserCardCode>
            <TransId><![CDATA[10022403432015000000000]]></TransId>
            <LocationId>291710000</LocationId>
            <Fee><![CDATA[10000]]></Fee>
            <OriginalFee><![CDATA[10000]]> </OriginalFee>
        </xml>
    """
    )
    assert message.target == "gh_e2243xxxxxxx"
    assert message.source == "oo2VNuOUuZGMxxxxxxxx"
    assert message.time == 1442390947
    assert message.type == "user_pay_from_pay_cell_event"
    assert message.card_id == "po2VNuCuRo-8sxxxxxxxxxxx"
    assert message.user_card_code == "38050000000"
    assert message.trans_id == "10022403432015000000000"
    assert message.location_id == 291710000
    assert message.fee == "10000"
    assert message.original_fee == "10000"


def test_user_view_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName> <![CDATA[gh_fcxxxx6a20993]]> </ToUserName>
            <FromUserName> <![CDATA[oZI8Fj040-xxxxx6gkoPOQTQ]]> </FromUserName>
            <CreateTime>1467811138</CreateTime>
            <MsgType> <![CDATA[event]]> </MsgType>
            <Event> <![CDATA[user_view_card]]> </Event>
            <CardId> <![CDATA[pZI8Fj2ezBbxxxxxT2UbiiWLb7Bg]]> </CardId>
            <UserCardCode> <![CDATA[4xxxxxxxx8558]]> </UserCardCode>
            <OuterStr> <![CDATA[12b]]> </OuterStr>
        </xml>
    """
    )
    assert message.target == "gh_fcxxxx6a20993"
    assert message.source == "oZI8Fj040-xxxxx6gkoPOQTQ"
    assert message.time == 1467811138
    assert message.type == "user_view_card_event"
    assert message.card_id == "pZI8Fj2ezBbxxxxxT2UbiiWLb7Bg"
    assert message.user_card_code == "4xxxxxxxx8558"
    assert message.outer_str == "12b"


def test_user_enter_session_from_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[toUser]]></ToUserName>
            <FromUserName><![CDATA[FromUser]]></FromUserName>
            <CreateTime>123456789</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[user_enter_session_from_card]]></Event>
            <CardId><![CDATA[cardid]]></CardId>
            <UserCardCode><![CDATA[12312312]]></UserCardCode>
        </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "user_enter_session_from_card_event"
    assert message.card_id == "cardid"
    assert message.user_card_code == "12312312"


def test_update_member_card_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[gh_9e1765b5568e]]></ToUserName>
            <FromUserName><![CDATA[ojZ8YtyVyr30HheH3CM73y7h4jJE]]></FromUserName>
            <CreateTime>1445507140</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[update_member_card]]></Event>
            <CardId><![CDATA[pjZ8Ytx-nwvpCRyQneH3Ncmh6N94]]></CardId>
            <UserCardCode><![CDATA[485027611252]]></UserCardCode>
            <ModifyBonus>3</ModifyBonus>
            <ModifyBalance>0</ModifyBalance>
        </xml>
    """
    )
    assert message.target == "gh_9e1765b5568e"
    assert message.source == "ojZ8YtyVyr30HheH3CM73y7h4jJE"
    assert message.time == 1445507140
    assert message.type == "update_member_card_event"
    assert message.card_id == "pjZ8Ytx-nwvpCRyQneH3Ncmh6N94"
    assert message.user_card_code == "485027611252"
    assert message.modify_bonus == 3
    assert message.modify_balance == 0


def test_card_sku_remind_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[gh_2d62d*****0]]></ToUserName>
            <FromUserName><![CDATA[oa3LFuBvWb7*********]]></FromUserName>
            <CreateTime>1443838506</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[card_sku_remind]]></Event>
            <CardId><![CDATA[pa3LFuAh2P65**********]]></CardId>
            <Detail><![CDATA[the card's quantity is equal to 0]]></Detail>
        </xml>
    """
    )
    assert message.target == "gh_2d62d*****0"
    assert message.source == "oa3LFuBvWb7*********"
    assert message.time == 1443838506
    assert message.type == "card_sku_remind_event"
    assert message.card_id == "pa3LFuAh2P65**********"
    assert message.detail == "the card's quantity is equal to 0"


def test_card_pay_order_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName><![CDATA[gh_7223c83d4be5]]></ToUserName>
            <FromUserName><![CDATA[ob5E7s-HoN9tslQY3-0I4qmgluHk]]></FromUserName>
            <CreateTime>1453295737</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[card_pay_order]]></Event>
            <OrderId><![CDATA[404091456]]></OrderId>
            <Status><![CDATA[ORDER_STATUS_FINANCE_SUCC]]></Status>
            <CreateOrderTime>1453295737</CreateOrderTime>
            <PayFinishTime>0</PayFinishTime>
            <Desc><![CDATA[]]></Desc>
            <FreeCoinCount><![CDATA[200]]></FreeCoinCount>
            <PayCoinCount><![CDATA[0]]></PayCoinCount>
            <RefundFreeCoinCount><![CDATA[0]]></RefundFreeCoinCount>
            <RefundPayCoinCount><![CDATA[0]]></RefundPayCoinCount>
            <OrderType><![CDATA[ORDER_TYPE_SYS_ADD]]></OrderType>
            <Memo><![CDATA[开通账户奖励]]></Memo>
            <ReceiptInfo><![CDATA[]]></ReceiptInfo>
        </xml>
    """
    )
    assert message.target == "gh_7223c83d4be5"
    assert message.source == "ob5E7s-HoN9tslQY3-0I4qmgluHk"
    assert message.time == 1453295737
    assert message.type == "card_pay_order_event"
    assert message.order_id == "404091456"
    assert message.status == "ORDER_STATUS_FINANCE_SUCC"
    assert message.create_order_time == 1453295737
    assert message.pay_finish_time == 0
    assert message.desc is None
    assert message.free_coin_count == "200"
    assert message.pay_coin_count == "0"
    assert message.refund_free_coin_count == "0"
    assert message.refund_pay_coin_count == "0"
    assert message.order_type == "ORDER_TYPE_SYS_ADD"
    assert message.memo == u"开通账户奖励"
    assert message.receipt_info is None


def test_submit_membercard_user_info_event():
    message = parse_user_msg(
        """
        <xml>
            <ToUserName> <![CDATA[gh_3fcea188bf78]]></ToUserName>
            <FromUserName><![CDATA[obLatjlaNQKb8FqOvt1M1x1lIBFE]]></FromUserName>
            <CreateTime>1432668700</CreateTime>
            <MsgType><![CDATA[event]]></MsgType>
            <Event><![CDATA[submit_membercard_user_info]]></Event>
            <CardId><![CDATA[pbLatjtZ7v1BG_ZnTjbW85GYc_E8]]></CardId>
            <UserCardCode><![CDATA[018255396048]]></UserCardCode>
        </xml>
    """
    )
    assert message.target == "gh_3fcea188bf78"
    assert message.source == "obLatjlaNQKb8FqOvt1M1x1lIBFE"
    assert message.time == 1432668700
    assert message.type == "submit_membercard_user_info_event"
    assert message.card_id == "pbLatjtZ7v1BG_ZnTjbW85GYc_E8"
    assert message.user_card_code == "018255396048"


def test_pic_weixin_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_e136c6e50636]]></ToUserName>
        <FromUserName><![CDATA[oMgHVjngRipVsoxg6TuX3vz6glDg]]></FromUserName>
        <CreateTime>1408090816</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[pic_weixin]]></Event>
        <EventKey><![CDATA[6]]></EventKey>
        <SendPicsInfo>
            <Count>1</Count>
            <PicList>
                <item>
                    <PicMd5Sum><![CDATA[5a75aaca956d97be686719218f275c6b]]></PicMd5Sum>
                </item>
            </PicList>
        </SendPicsInfo>
    </xml>
    """
    )
    assert message.target == "gh_e136c6e50636"
    assert message.source == "oMgHVjngRipVsoxg6TuX3vz6glDg"
    assert message.time == 1408090816
    assert message.type == "pic_weixin_event"
    assert message.key == "6"
    assert message.count == 1
    assert message.pic_list == [
        {
            'pic_md5_sum': '5a75aaca956d97be686719218f275c6b'
        }
    ]


def test_location_select_event():
    message = parse_user_msg(
        u"""
    <xml>
        <ToUserName><![CDATA[gh_e136c6e50636]]></ToUserName>
        <FromUserName><![CDATA[oMgHVjngRipVsoxg6TuX3vz6glDg]]></FromUserName>
        <CreateTime>1408091189</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[location_select]]></Event>
        <EventKey><![CDATA[6]]></EventKey>
        <SendLocationInfo>
            <Location_X><![CDATA[23]]></Location_X>
            <Location_Y><![CDATA[113]]></Location_Y>
            <Scale><![CDATA[15]]></Scale>
            <Label><![CDATA[广州市海珠区客村艺苑路 106号]]></Label>
            <Poiname><![CDATA[]]></Poiname>
        </SendLocationInfo>
    </xml>
    """
    )
    assert message.target == "gh_e136c6e50636"
    assert message.source == "oMgHVjngRipVsoxg6TuX3vz6glDg"
    assert message.time == 1408091189
    assert message.type == "location_select_event"
    assert message.key == "6"
    assert message.location_x == "23"
    assert message.location_y == "113"
    assert message.scale == "15"
    assert message.label == u"广州市海珠区客村艺苑路 106号"
    assert message.poi_name is None


def test_click_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[CLICK]]></Event>
        <EventKey><![CDATA[EVENTKEY]]></EventKey>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 123456789
    assert message.type == "click_event"
    assert message.key == "EVENTKEY"


def test_view_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[VIEW]]></Event>
        <EventKey><![CDATA[www.qq.com]]></EventKey>
    </xml>"""
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "view_event"
    assert message.key == "www.qq.com"


def test_location_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[fromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[LOCATION]]></Event>
        <Latitude>23.137466</Latitude>
        <Longitude>113.352425</Longitude>
        <Precision>119.385040</Precision>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "fromUser"
    assert message.time == 123456789
    assert message.type == "location_event"
    assert message.latitude == 23.137466
    assert message.longitude == 113.352425
    assert message.precision == 119.385040


def test_templatesendjobfinish_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_7f083739789a]]></ToUserName>
        <FromUserName><![CDATA[oia2TjuEGTNoeX76QEjQNrcURxG8]]></FromUserName>
        <CreateTime>1395658920</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[TEMPLATESENDJOBFINISH]]></Event>
        <MsgID>200163836</MsgID>
        <Status><![CDATA[success]]></Status>
    </xml>
    """
    )
    assert message.message_id == 200163836
    assert message.status == 'success'

    assert parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_7f083739789a]]></ToUserName>
        <FromUserName><![CDATA[oia2TjuEGTNoeX76QEjQNrcURxG8]]></FromUserName>
        <CreateTime>1395658984</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[TEMPLATESENDJOBFINISH]]></Event>
        <MsgID>200163840</MsgID>
        <Status><![CDATA[failed: system failed]]></Status>
    </xml>
    """
    ).status == 'failed: system failed'


def test_user_scan_product_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_4e47c9c9ecad]]></ToUserName>
        <FromUserName><![CDATA[okkeXs1nI-xU4ql8-5BXkv1f0gDo]]></FromUserName>
        <CreateTime>1438250110</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[user_scan_product]]></Event>
        <KeyStandard><![CDATA[ean13]]></KeyStandard>
        <KeyStr><![CDATA[6901481811083]]></KeyStr>
        <Country><![CDATA[中国]]></Country>
        <Province><![CDATA[广东]]></Province>
        <City><![CDATA[揭阳]]></City>
        <Sex>1</Sex>
        <Scene>2</Scene>
        <ExtInfo><![CDATA[123]]></ExtInfo>
    </xml>
    """
    )
    assert message.target == "gh_4e47c9c9ecad"
    assert message.source == "okkeXs1nI-xU4ql8-5BXkv1f0gDo"
    assert message.time == 1438250110
    assert message.type == "user_scan_product_event"
    assert message.key_standard == "ean13"
    assert message.key_str == "6901481811083"
    assert message.country == u"中国"
    assert message.province == u"广东"
    assert message.city == u"揭阳"
    assert message.sex == 1
    assert message.scene == 2
    assert message.ext_info == "123"


def test_user_scan_product_enter_session_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_fbe8a958756e]]></ToUserName>
        <FromUserName><![CDATA[otAzGjrS4AYCmeJM1GhEOcHXXTAo]]></FromUserName>
        <CreateTime>1433259128</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[user_scan_product_enter_session]]></Event>
        <KeyStandard><![CDATA[ena13]]></KeyStandard>
        <KeyStr><![CDATA[6954767461373]]></KeyStr>
        <ExtInfo><![CDATA[]]></ExtInfo>
    </xml>
    """
    )
    assert message.target == "gh_fbe8a958756e"
    assert message.source == "otAzGjrS4AYCmeJM1GhEOcHXXTAo"
    assert message.time == 1433259128
    assert message.type == "user_scan_product_enter_session_event"
    assert message.key_standard == "ena13"
    assert message.key_str == "6954767461373"
    assert message.ext_info is None


def test_user_scan_product_async_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_fbe8a958756e]]></ToUserName>
        <FromUserName><![CDATA[otAzGjrS4AYCmeJM1GhEOcHXXTAo]]></FromUserName>
        <CreateTime>1434541327</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[user_scan_product_async]]></Event>
        <KeyStandard><![CDATA[qrcode]]></KeyStandard>
        <KeyStr><![CDATA[lincolntest2]]></KeyStr>
        <ExtInfo><![CDATA[123]]></ExtInfo>
        <RegionCode><![CDATA[440105]]></RegionCode>
    </xml>
    """
    )
    assert message.target == "gh_fbe8a958756e"
    assert message.source == "otAzGjrS4AYCmeJM1GhEOcHXXTAo"
    assert message.time == 1434541327
    assert message.type == "user_scan_product_async_event"
    assert message.key_standard == "qrcode"
    assert message.key_str == "lincolntest2"
    assert message.ext_info == "123"
    assert message.region_code == "440105"


def test_user_scan_product_verify_action_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[gh_404e58ec528e]]></ToUserName>
        <FromUserName><![CDATA[od_ikt8qi21-hVTtYgm8xSfTLH5w]]></FromUserName>
        <CreateTime>1450429257</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[user_scan_product_verify_action]]></Event>
        <KeyStandard><![CDATA[ean13]]></KeyStandard>
        <KeyStr><![CDATA[6901481811083]]></KeyStr>
        <Result><![CDATA[verify_ok]]></Result>
        <ReasonMsg><![CDATA[]]></ReasonMsg>
    </xml>
    """
    )
    assert message.target == "gh_404e58ec528e"
    assert message.source == "od_ikt8qi21-hVTtYgm8xSfTLH5w"
    assert message.time == 1450429257
    assert message.type == "user_scan_product_verify_action_event"
    assert message.key_standard == "ean13"
    assert message.key_str == "6901481811083"
    assert message.result == "verify_ok"
    assert message.reason_msg is None


def test_unknown_event():
    message = parse_user_msg(
        """
    <xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[FromUser]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[unknown]]></Event>
    </xml>
    """
    )
    assert message.target == "toUser"
    assert message.source == "FromUser"
    assert message.time == 123456789
    assert message.type == "unknown_event"



# -*- coding: utf-8 -*-

import time
import requests
import urllib.parse

from requests.compat import json as _json
from zgrobot.utils import to_text
from zgrobot.replies import Article


class ClientException(Exception):
    pass


def check_error(json):
    """
    检测微信公众平台返回值中是否包含错误的返回码。
    如果返回码提示有错误，抛出一个 :class:`ClientException` 异常。否则返回 True 。
    """
    if "errcode" in json and json["errcode"] != 0:
        raise ClientException("{}: {}".format(json["errcode"], json["errmsg"]))
    return json


def _build_send_data(msg_type, content):
    """
    编译群发参数
    :param msg_type: 群发类型，图文消息为 mpnews，文本消息为 text，语音为 voice，音乐为 music，图片为 image，视频为 video，卡券为 wxcard。
    :param content: 群发内容。
    :return: 群发参数。
    """
    send_data = {}
    send_data['msgtype'] = msg_type
    if msg_type in ['mpnews', 'voice', 'music', 'image']:
        send_data[msg_type] = {'media_id': content}
    elif msg_type == 'video':
        send_data['mpvideo'] = {'media_id': content}
        send_data['msgtype'] = 'mpvideo'
    elif msg_type == 'text':
        send_data['text'] = {'content': content}
    elif msg_type == 'wxcard':
        send_data['wxcard'] = {'card_id': content}
    else:
        send_data['text'] = {'content': content}
        send_data['msgtype'] = 'text'
    return send_data


class Client(object):
    """
    微信 API 操作类
    通过这个类可以方便的通过微信 API 进行一系列操作，比如主动发送消息、创建自定义菜单等
    """
    def __init__(self, config):
        self.config = config
        self._token = None
        self.token_expires_at = None

    @property
    def appid(self):
        return self.config.get("APP_ID", None)

    @property
    def appsecret(self):
        return self.config.get("APP_SECRET", None)

    @staticmethod
    def _url_encode_files(file):
        if hasattr(file, "name"):
            file = (urllib.parse.quote(file.name), file)
        return file

    def request(self, method, url, **kwargs):
        if "params" not in kwargs:
            kwargs["params"] = {"access_token": self.token}
        if isinstance(kwargs.get("data", ""), dict):
            body = _json.dumps(kwargs["data"], ensure_ascii=False)
            body = body.encode('utf8')
            kwargs["data"] = body

        r = requests.request(method=method, url=url, **kwargs)
        r.raise_for_status()
        r.encoding = "utf-8"
        json = r.json()
        if check_error(json):
            return json

    def get(self, url, **kwargs):
        return self.request(method="get", url=url, **kwargs)

    def post(self, url, **kwargs):
        if "files" in kwargs:
            # Although there is only one key "media" possible in "files" now,
            # we decide to check every key to support possible keys in the future
            # Fix chinese file name error #292
            kwargs["files"] = dict(
                zip(
                    kwargs["files"],
                    map(self._url_encode_files, kwargs["files"].values())
                )
            )
        return self.request(method="post", url=url, **kwargs)

    def grant_token(self):
        """
        获取 Access Token。

        :return: 返回的 ``JSON`` 数据包

                正常时返回 ``{"access_token":"ACCESS_TOKEN", "expires_in":7200}``

                错误时返回 ``{"errcode":40013,"errmsg":"invalid appid"}``
        """
        return self.get(
            url="https://api.weixin.qq.com/cgi-bin/token",
            params={
                "grant_type": "client_credential",
                "appid": self.appid,
                "secret": self.appsecret
            }
        )

    def get_access_token(self) -> str:
        """
        判断现有的token是否过期。
        用户需要多进程或者多机部署可以手动重写这个函数
        来自定义token的存储，刷新策略。

        :return: 返回 ``token``
        """
        if self._token:
            now = time.time()
            if self.token_expires_at - now > 60:
                return self._token
        json_str = self.grant_token()
        self._token = json_str["access_token"]
        self.token_expires_at = int(time.time()) + json_str["expires_in"]
        return self._token

    @property
    def token(self):
        return self.get_access_token()

    def get_ip_list(self):
        """
        获取微信服务器IP地址。

        :return: 返回的 JSON 数据包

                正常时返回 ``{"ip_list": ["127.0.0.1", "127.0.0.2", "101.226.103.0/25"]}``

                错误时返回 ``{"errcode":40013,"errmsg":"invalid appid"}``
        """
        return self.get(url="https://api.weixin.qq.com/cgi-bin/getcallbackip")

    def create_menu(self, menu_data):
        """
        创建自定义菜单::

            client.create_menu({
                "button":[
                    {
                        "type":"click",
                        "name":"今日歌曲",
                        "key":"V1001_TODAY_MUSIC"
                    },
                    {
                        "type":"click",
                        "name":"歌手简介",
                        "key":"V1001_TODAY_SINGER"
                    },
                    {
                        "name":"菜单",
                        "sub_button":[
                            {
                                "type":"view",
                                "name":"搜索",
                                "url":"http://www.soso.com/"
                            },
                            {
                                "type":"view",
                                "name":"视频",
                                "url":"http://v.qq.com/"
                            },
                            {
                                "type":"click",
                                "name":"赞一下我们",
                                "key":"V1001_GOOD"
                            }
                        ]
                    }
                ]})

        :param menu_data: Python 字典
        :return: 返回的 JSON 数据包

                正确时返回 ``{"errcode":0,"errmsg":"ok"}``

                错误时返回 ``{"errcode":40018,"errmsg":"invalid button name size"}``
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/menu/create",
            data=menu_data
        )

    def get_menu(self):
        """
        查询自定义菜单。

        :return: 返回的 JSON 数据包
        """
        return self.get("https://api.weixin.qq.com/cgi-bin/menu/get")

    def delete_menu(self):
        """
        删除自定义菜单。

        :return: 返回的 JSON 数据包

                正确时返回 ``{"errcode":0,"errmsg":"ok"}``
        """
        return self.get("https://api.weixin.qq.com/cgi-bin/menu/delete")

    def create_custom_menu(self, menu_data, matchrule):
        """
        创建个性化菜单::

            button = [
                {
                    "type":"click",
                    "name":"今日歌曲",
                    "key":"V1001_TODAY_MUSIC"
                },
                {
                    "name":"菜单",
                    "sub_button":[
                    {
                        "type":"view",
                        "name":"搜索",
                        "url":"http://www.soso.com/"
                    },
                    {
                        "type":"view",
                        "name":"视频",
                        "url":"http://v.qq.com/"
                    },
                    {
                        "type":"click",
                        "name":"赞一下我们",
                        "key":"V1001_GOOD"
                    }]
             }]
             matchrule = {
                "group_id":"2",
                "sex":"1",
                "country":"中国",
                "province":"广东",
                "city":"广州",
                "client_platform_type":"2",
                "language":"zh_CN"
            }
            client.create_custom_menu(button, matchrule)

        :param menu_data: 如上所示的 Python 字典
        :param matchrule: 如上所示的匹配规则
        :return: 返回的 JSON 数据包

                正确时返回 ``{"menuid":"208379533"}``

                错误时的返回码请见接口返回码说明.
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/menu/addconditional",
            data={
                "button": menu_data,
                "matchrule": matchrule
            }
        )

    def delete_custom_menu(self, menu_id):
        """
        删除个性化菜单。

        :param menu_id: 菜单的 ID
        :return: 返回的 JSON 数据包

                正确时返回 ``{"errcode":0,"errmsg":"ok"}``

                错误时的返回码请见接口返回码说明。
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/menu/delconditional",
            data={"menuid": menu_id}
        )

    def match_custom_menu(self, user_id):
        """
        测试个性化菜单匹配结果。

        :param user_id: 要测试匹配的用户 ID，也可以是用户的微信号
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/menu/trymatch",
            data={"user_id": user_id}
        )

    def get_custom_menu_config(self):
        """
        获取自定义菜单配置接口。

        :return: 返回的 JSON 数据包
        """
        return self.get(
            url="https://api.weixin.qq.com/cgi-bin/get_current_selfmenu_info"
        )

    def add_custom_service_account(self, account, nickname, password):
        """
        添加客服帐号。

        :param account: 客服账号的用户名
        :param nickname: 客服账号的昵称
        :param password: 客服账号的密码
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/customservice/kfaccount/add",
            data={
                "kf_account": account,
                "nickname": nickname,
                "password": password
            }
        )

    def update_custom_service_account(self, account, nickname, password):
        """
        修改客服帐号。

        :param account: 客服账号的用户名
        :param nickname: 客服账号的昵称
        :param password: 客服账号的密码
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/customservice/kfaccount/update",
            data={
                "kf_account": account,
                "nickname": nickname,
                "password": password
            }
        )

    def delete_custom_service_account(self, account, nickname, password):
        """
        删除客服帐号。

        :param account: 客服账号的用户名
        :param nickname: 客服账号的昵称
        :param password: 客服账号的密码
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/customservice/kfaccount/del",
            data={
                "kf_account": account,
                "nickname": nickname,
                "password": password
            }
        )

    def upload_custom_service_account_avatar(self, account, avatar):
        """
        设置客服帐号的头像。

        :param account: 客服账号的用户名
        :param avatar: 头像文件，必须是 jpg 格式
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url=
            "http://api.weixin.qq.com/customservice/kfaccount/uploadheadimg",
            params={
                "access_token": self.token,
                "kf_account": account
            },
            files={"media": avatar}
        )

    def get_custom_service_account_list(self):
        """
        获取所有客服账号。

        :return: 返回的 JSON 数据包
        """
        return self.get(
            url="https://api.weixin.qq.com/cgi-bin/customservice/getkflist"
        )

    def get_online_custom_service_account_list(self):
        """
        获取状态为"在线"的客服账号列表。

        :return: 返回的 JSON 数据包
        """
        return self.get(
            url="https://api.weixin.qq.com/cgi-bin/customservice/getonlinekflist"
        )

    def upload_media(self, media_type, media_file):
        """
        上传临时多媒体文件。

        :param media_type: 媒体文件类型，分别有图片（image）、语音（voice）、视频（video）和缩略图（thumb）
        :param media_file: 要上传的文件，一个 File-object
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/media/upload",
            params={
                "access_token": self.token,
                "type": media_type
            },
            files={"media": media_file}
        )

    def download_media(self, media_id):
        """
        下载临时多媒体文件。

        :param media_id: 媒体文件 ID
        :return: requests 的 Response 实例
        """
        return requests.get(
            url="https://api.weixin.qq.com/cgi-bin/media/get",
            params={
                "access_token": self.token,
                "media_id": media_id
            }
        )

    def add_news(self, articles):
        """
        新增永久图文素材::

            articles = [{
               "title": TITLE,
               "thumb_media_id": THUMB_MEDIA_ID,
               "author": AUTHOR,
               "digest": DIGEST,
               "show_cover_pic": SHOW_COVER_PIC(0 / 1),
               "content": CONTENT,
               "content_source_url": CONTENT_SOURCE_URL
            }
            # 若新增的是多图文素材，则此处应有几段articles结构，最多8段
            ]
            client.add_news(articles)

        :param articles: 如示例中的数组
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/material/add_news",
            data={"articles": articles}
        )

    def upload_news_picture(self, file):
        """
        上传图文消息内的图片。
        TODO 修改参数 file: File-object -> 文件路径/文件 url

        :param file: 要上传的文件，一个 File-object
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/media/uploadimg",
            params={"access_token": self.token},
            files={"media": file}
        )

    def upload_permanent_media(self, media_type, media_file):
        """
        上传其他类型永久素材。
        TODO 修改参数 media_file: File-object -> 文件路径/文件 url

        :param media_type: 媒体文件类型，分别有图片（image）、语音（voice）和缩略图（thumb）
        :param media_file: 要上传的文件，一个 File-object
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/material/add_material",
            params={
                "access_token": self.token,
                "type": media_type
            },
            files={"media": media_file}
        )

    def upload_permanent_video(self, title, introduction, video):
        """
        上传永久视频。
        TODO 修改参数 video: File-object -> 文件路径/文件 url

        :param title: 视频素材的标题
        :param introduction: 视频素材的描述
        :param video: 要上传的视频，一个 File-object
        :return: requests 的 Response 实例
        """
        return requests.post(
            url="https://api.weixin.qq.com/cgi-bin/material/add_material",
            params={
                "access_token": self.token,
                "type": "video"
            },
            data={
                "description": _json.dumps(
                    {
                        "title": title,
                        "introduction": introduction
                    },
                    ensure_ascii=False
                ).encode("utf-8")
            },
            files={"media": video}
        )

    def download_permanent_media(self, media_id):
        """
        获取永久素材。

        :param media_id: 媒体文件 ID
        :return: requests 的 Response 实例
        """
        return requests.post(
            url="https://api.weixin.qq.com/cgi-bin/material/get_material",
            params={"access_token": self.token},
            data=_json.dumps({
                "media_id": media_id
            }, ensure_ascii=False).encode("utf-8")
        )

    def delete_permanent_media(self, media_id):
        """
        删除永久素材。

        :param media_id: 媒体文件 ID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/material/del_material",
            data={"media_id": media_id}
        )

    def update_news(self, update_data):
        """
        修改永久图文素材::

            update_data = {
                "media_id":MEDIA_ID,
                "index":INDEX,
                "articles": {
                    "title": TITLE,
                    "thumb_media_id": THUMB_MEDIA_ID,
                    "author": AUTHOR,
                    "digest": DIGEST,
                    "show_cover_pic": SHOW_COVER_PIC(0 / 1),
                    "content": CONTENT,
                    "content_source_url": CONTENT_SOURCE_URL
                }
            }
            client.update_news(update_data)

        :param update_data: 更新的数据，要包含 media_id（图文素材的 ID），index（要更新的文章在图文消息中的位置），articles（新的图文素材数据）
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/material/update_news",
            data=update_data
        )

    def get_media_count(self):
        """
        获取素材总数。

        :return: 返回的 JSON 数据包
        """
        return self.get(
            url="https://api.weixin.qq.com/cgi-bin/material/get_materialcount"
        )

    def get_media_list(self, media_type, offset, count):
        """
        获取素材列表。

        :param media_type: 素材的类型，图片（image）、视频（video）、语音 （voice）、图文（news）
        :param offset: 从全部素材的该偏移位置开始返回，0表示从第一个素材返回
        :param count: 返回素材的数量，取值在1到20之间
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/material/batchget_material",
            data={
                "type": media_type,
                "offset": offset,
                "count": count
            }
        )

    def create_group(self, name):
        """
        创建分组。

        :param name: 分组名字（30个字符以内）
        :return: 返回的 JSON 数据包

        """
        name = to_text(name)
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/groups/create",
            data={"group": {
                "name": name
            }}
        )

    def get_groups(self):
        """
        查询所有分组。

        :return: 返回的 JSON 数据包
        """
        return self.get("https://api.weixin.qq.com/cgi-bin/groups/get")

    def get_group_by_id(self, openid):
        """
        查询用户所在分组。

        :param openid: 用户的OpenID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/groups/getid",
            data={"openid": openid}
        )

    def update_group(self, group_id, name):
        """
        修改分组名。

        :param group_id: 分组 ID，由微信分配
        :param name: 分组名字（30个字符以内）
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/groups/update",
            data={"group": {
                "id": int(group_id),
                "name": to_text(name)
            }}
        )

    def move_user(self, user_id, group_id):
        """
        移动用户分组。

        :param user_id: 用户 ID，即收到的 `Message` 的 source
        :param group_id: 分组 ID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/groups/members/update",
            data={
                "openid": user_id,
                "to_groupid": group_id
            }
        )

    def move_users(self, user_id_list, group_id):
        """
        批量移动用户分组。

        :param user_id_list: 用户 ID 的列表（长度不能超过50）
        :param group_id: 分组 ID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/groups/members/batchupdate",
            data={
                "openid_list": user_id_list,
                "to_groupid": group_id
            }
        )

    def delete_group(self, group_id):
        """
        删除分组。

        :param group_id: 要删除的分组的 ID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/groups/delete",
            data={"group": {
                "id": group_id
            }}
        )

    def remark_user(self, user_id, remark):
        """
        设置备注名。

        :param user_id: 设置备注名的用户 ID
        :param remark: 新的备注名，长度必须小于30字符
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/user/info/updateremark",
            data={
                "openid": user_id,
                "remark": remark
            }
        )

    def get_user_info(self, user_id, lang="zh_CN"):
        """
        获取用户基本信息。

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param lang: 返回国家地区语言版本，zh_CN 简体，zh_TW 繁体，en 英语
        :return: 返回的 JSON 数据包
        """
        return self.get(
            url="https://api.weixin.qq.com/cgi-bin/user/info",
            params={
                "access_token": self.token,
                "openid": user_id,
                "lang": lang
            }
        )

    def get_users_info(self, user_id_list, lang="zh_CN"):
        """
        批量获取用户基本信息。

        :param user_id_list: 用户 ID 的列表
        :param lang: 返回国家地区语言版本，zh_CN 简体，zh_TW 繁体，en 英语
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/user/info/batchget",
            data={
                "user_list": [
                    {
                        "openid": user_id,
                        "lang": lang
                    } for user_id in user_id_list
                ]
            }
        )

    def get_followers(self, first_user_id=None):
        """
        获取关注者列表
        详情请参考 https://developers.weixin.qq.com/doc/offiaccount/User_Management/Getting_a_User_List.html

        :param first_user_id: 可选。第一个拉取的OPENID，不填默认从头开始拉取
        :return: 返回的 JSON 数据包
        """
        params = {"access_token": self.token}
        if first_user_id:
            params["next_openid"] = first_user_id
        return self.get(
            "https://api.weixin.qq.com/cgi-bin/user/get", params=params
        )

    def send_text_message(self, user_id, content, kf_account=None):
        """
        发送文本消息。

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param content: 消息正文
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        data = {
            "touser": user_id,
            "msgtype": "text",
            "text": {
                "content": content
            }
        }
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_image_message(self, user_id, media_id, kf_account=None):
        """
        发送图片消息。

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param media_id: 图片的媒体ID。 可以通过 :func:`upload_media` 上传。
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        data = {
            "touser": user_id,
            "msgtype": "image",
            "image": {
                "media_id": media_id
            }
        }
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_voice_message(self, user_id, media_id, kf_account=None):
        """
        发送语音消息。

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param media_id: 发送的语音的媒体ID。 可以通过 :func:`upload_media` 上传。
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        data = {
            "touser": user_id,
            "msgtype": "voice",
            "voice": {
                "media_id": media_id
            }
        }
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_video_message(
        self,
        user_id,
        media_id,
        title=None,
        description=None,
        kf_account=None
    ):
        """
        发送视频消息。

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param media_id: 发送的视频的媒体ID。 可以通过 :func:`upload_media` 上传。
        :param title: 视频消息的标题
        :param description: 视频消息的描述
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        video_data = {
            "media_id": media_id,
        }
        if title:
            video_data["title"] = title
        if description:
            video_data["description"] = description
        data = {"touser": user_id, "msgtype": "video", "video": video_data}
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_music_message(
        self,
        user_id,
        url,
        hq_url,
        thumb_media_id,
        title=None,
        description=None,
        kf_account=None
    ):
        """
        发送音乐消息。
        注意如果你遇到了缩略图不能正常显示的问题， 不要慌张； 目前来看是微信服务器端的问题。
        对此我们也无能为力 ( `#197 <https://github.com/whtsky/ZgRoBot/issues/197>`_ )

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param url: 音乐链接
        :param hq_url: 高品质音乐链接，wifi环境优先使用该链接播放音乐
        :param thumb_media_id: 缩略图的媒体ID。 可以通过 :func:`upload_media` 上传。
        :param title: 音乐标题
        :param description: 音乐描述
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        music_data = {
            "musicurl": url,
            "hqmusicurl": hq_url,
            "thumb_media_id": thumb_media_id
        }
        if title:
            music_data["title"] = title
        if description:
            music_data["description"] = description
        data = {"touser": user_id, "msgtype": "music", "music": music_data}
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}

        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_article_message(self, user_id, articles, kf_account=None):
        """
        发送图文消息::

            articles = [
                {
                    "title":"Happy Day",
                    "description":"Is Really A Happy Day",
                    "url":"URL",
                    "picurl":"PIC_URL"
                },
                {
                    "title":"Happy Day",
                    "description":"Is Really A Happy Day",
                    "url":"URL",
                    "picurl":"PIC_URL"
                }
            ]
            client.send_acticle_message("user_id", acticles)

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param articles: 一个包含至多8个 article 字典或 Article 对象的数组
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        if isinstance(articles[0], Article):
            formatted_articles = []
            for article in articles:
                result = article.args
                result["picurl"] = result.pop("img")
                formatted_articles.append(result)
        else:
            formatted_articles = articles
        data = {
            "touser": user_id,
            "msgtype": "news",
            "news": {
                "articles": formatted_articles
            }
        }
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_news_message(self, user_id, media_id, kf_account=None):
        """
        发送永久素材中的图文消息。

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param media_id: 媒体文件 ID
        :param kf_account: 发送消息的客服账户，默认值为 None，None 为不指定
        :return: 返回的 JSON 数据包
        """
        data = {
            "touser": user_id,
            "msgtype": "mpnews",
            "mpnews": {
                "media_id": media_id
            }
        }
        if kf_account is not None:
            data['customservice'] = {'kf_account': kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def send_miniprogrampage_message(
        self,
        user_id,
        title,
        appid,
        pagepath,
        thumb_media_id,
        kf_account=None
    ):
        """
        发送小程序卡片（要求小程序与公众号已关联）

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param title: 小程序卡片的标题
        :param appid: 小程序的 appid，要求小程序的 appid 需要与公众号有关联关系
        :param pagepath: 小程序的页面路径，跟 app.json 对齐，支持参数，比如 pages/index/index?foo=bar
        :param thumb_media_id: 小程序卡片图片的媒体 ID，小程序卡片图片建议大小为 520*416
        :param kf_account: 需要以某个客服帐号来发消息时指定的客服账户
        :return: 返回的 JSON 数据包
        """
        data = {
            "touser": user_id,
            "msgtype": "miniprogrampage",
            "miniprogrampage": {
                "title": title,
                "appid": appid,
                "pagepath": pagepath,
                "thumb_media_id": thumb_media_id
            }
        }
        if kf_account is not None:
            data["customservice"] = {"kf_account": kf_account}
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/custom/send",
            data=data
        )

    def create_qrcode(self, data):
        """
        创建二维码。

        :param data: 你要发送的参数 dict
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/qrcode/create", data=data
        )

    def show_qrcode(self, ticket):
        """
        通过ticket换取二维码。

        :param ticket: 二维码 ticket 。可以通过 :func:`create_qrcode` 获取到
        :return: 返回的 Request 对象
        """
        return requests.get(
            url="https://mp.weixin.qq.com/cgi-bin/showqrcode",
            params={"ticket": ticket}
        )

    def send_template_message(
        self, user_id, template_id, data, url='', miniprogram=None
    ):
        """
        发送模板消息
        详情请参考 https://developers.weixin.qq.com/doc/offiaccount/Message_Management/Template_Message_Interface.html

        :param user_id: 用户 ID 。 就是你收到的 `Message` 的 source
        :param template_id: 模板 ID。
        :param data: 用于渲染模板的数据。
        :param url: 模板消息的可选链接。
        :param miniprogram: 跳小程序所需数据的可选数据。
        :return: 返回的 JSON 数据包
        """
        payload = {
            "touser": user_id,
            "template_id": template_id,
            "url": url,
            "data": data
        }
        if miniprogram:
            payload["miniprogram"] = miniprogram
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/template/send",
            data=payload
        )

    def create_tag(self, tag_name):
        """
        创建一个新标签

        :param tag_name: 标签名
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/tags/create",
            data={"tag": {
                "name": tag_name
            }}
        )

    def get_tags(self):
        """
        获取已经存在的标签

        :return: 返回的 JSON 数据包
        """
        return self.get(url="https://api.weixin.qq.com/cgi-bin/tags/get", )

    def update_tag(self, tag_id, tag_name):
        """
        修改标签

        :param tag_id: 标签 ID
        :param tag_name: 新的标签名
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/tags/update",
            data={"tag": {
                "id": tag_id,
                "name": tag_name
            }}
        )

    def delete_tag(self, tag_id):
        """
        删除标签

        :param tag_id: 标签 ID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/tags/delete",
            data={"tag": {
                "id": tag_id,
            }}
        )

    def get_users_by_tag(self, tag_id, next_open_id=""):
        """
        获取标签下粉丝列表

        :param tag_id: 标签 ID
        :param next_open_id: 第一个拉取用户的 OPENID，默认从头开始拉取
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/user/tag/get",
            data={
                "tagid": tag_id,
                "next_openid": next_open_id
            }
        )

    def get_tags_by_user(self, open_id):
        """
        获取用户身上的标签列表

        :param open_id: 用户的 OPENID
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/tags/getidlist",
            data={
                "openid": open_id,
            }
        )

    def tag_users(self, tag_id, open_id_list):
        """
        批量为用户打标签

        :param tag_id: 标签 ID
        :param open_id_list: 包含一个或多个用户的 OPENID 的列表
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/tags/members/batchtagging",
            data={
                "openid_list": open_id_list,
                "tagid": tag_id
            }
        )

    def untag_users(self, tag_id, open_id_list):
        """
        批量为用户取消标签

        :param tag_id: 标签 ID
        :param open_id_list: 包含一个或多个用户的 OPENID 的列表
        :return: 返回的 JSON 数据包
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/tags/members/batchuntagging",
            data={
                "openid_list": open_id_list,
                "tagid": tag_id
            }
        )

    def upload_news(self, articles):
        """
        上传图文消息素材 ::

            articles = [{
                "thumb_media_id":"qI6_Ze_6PtV7svjolgs-rN6stStuHIjs9_DidOHaj0Q-mwvBelOXCFZiq2OsIU-p",
                "author":"xxx",
                "title":"Happy Day",
                "content_source_url":"www.qq.com",
                "content":"content",
                "digest":"digest",
                "show_cover_pic":1,
                "need_open_comment":1,
                "only_fans_can_comment":1
            }]

        具体请参考: https://developers.weixin.qq.com/doc/offiaccount/Message_Management/Batch_Sends_and_Originality_Checks.html#1


        :param articles: 上传的图文消息数据。
        :return: 返回的 JSON 数据包。
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/media/uploadnews",
            data={
                "articles": articles,
            }
        )

    def send_mass_msg(
        self,
        msg_type,
        content,
        user_list=None,
        send_ignore_reprint=False,
        client_msg_id=None
    ):
        """
        向指定对象群发信息。

        :param msg_type: 群发类型，图文消息为 mpnews，文本消息为 text，语音为 voice，音乐为 music，图片为 image，视频为 video，卡券为 wxcard。
        :param content: 群发内容。
        :param user_list: 发送对象，整型代表用户组，列表代表指定用户，如果为 None 则代表全部发送。
        :param send_ignore_reprint: 图文消息被判定为转载时，是否继续群发。 True 为继续群发（转载），False 为停止群发。 该参数默认为 False。
        :param client_msg_id: 群发时，微信后台将对 24 小时内的群发记录进行检查，如果该 clientmsgid 已经存在一条群发记录，则会拒绝本次群发请求，返回已存在的群发 msgid, 控制再 64 个字符内。
        :return: 返回的 JSON 数据包。
        """
        send_data = _build_send_data(msg_type, content)
        send_data['send_ignore_reprint'] = send_ignore_reprint
        if client_msg_id is not None:
            send_data['clientmsgid'] = client_msg_id
        if isinstance(user_list, list):
            url = 'https://api.weixin.qq.com/cgi-bin/message/mass/send'
            send_data['touser'] = user_list
        else:
            url = 'https://api.weixin.qq.com/cgi-bin/message/mass/sendall'
            if user_list == None:
                send_data['filter'] = {
                    "is_to_all": True,
                }
            else:
                send_data['filter'] = {"is_to_all": False, "tag_id": user_list}

        return self.post(url=url, data=send_data)

    def delete_mass_msg(self, msg_id, article_idx=0):
        """
        群发之后，随时可以通过该接口删除群发。

        :param msg_id: 发送出去的消息 ID。
        :param article_idx: 要删除的文章在图文消息中的位置，第一篇编号为 1，该字段不填或填 0 会删除全部文章。
        :return: 微信返回的 json 数据。
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/mass/delete",
            data={
                "msg_id": msg_id,
                "article_idx": article_idx
            }
        )

    def send_mass_preview_to_user(
        self, msg_type, content, user, user_type='openid'
    ):
        """
        开发者可通过该接口发送消息给指定用户，在手机端查看消息的样式和排版。为了满足第三方平台开发者的需求，在保留对 openID 预览能力的同时，增加了对指定微信号发送预览的能力，但该能力每日调用次数有限制（100 次），请勿滥用。

        :param user_type: 预览对象，`openid` 代表以 openid 发送，`wxname` 代表以微信号发送。
        :param msg_type: 发送类型，图文消息为 mpnews，文本消息为 text，语音为 voice，音乐为 music，图片为 image，视频为 video，卡券为 wxcard。
        :param content: 预览内容。
        :param user: 预览用户。
        :return: 返回的 json。
        """
        send_data = _build_send_data(msg_type, content)
        if user_type == 'openid':
            send_data['touser'] = user
        else:
            send_data['towxname'] = user
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/mass/preview",
            data=send_data
        )

    def get_mass_msg_status(self, msg_id):
        """
        查询群发消息发送状态。

        :param msg_id: 群发消息后返回的消息 id。
        :return: 返回的 json。
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/mass/get",
            data={'msg_id': msg_id}
        )

    def get_mass_msg_speed(self):
        """
        获取群发速度。

        :return: 返回的 json。
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/mass/speed/get",
            data={}
        )

    def set_mass_msg_speed(self, speed):
        """
        设置群发速度。

        :param speed: 群发速度的级别，是一个 0 到 4 的整数，数字越大表示群发速度越慢。
        :return: 返回的 json。
        """
        return self.post(
            url="https://api.weixin.qq.com/cgi-bin/message/mass/speed/set",
            data={"speed": speed}
        )



# -*- coding: utf-8 -*-

import io
import json
import os
import random
import re
import string
import time
from functools import wraps
from hashlib import sha1

try:
    from secrets import choice
except ImportError:
    from random import choice

string_types = (str, bytes)

re_type = type(re.compile("regex_test"))


def get_signature(token, timestamp, nonce, *args):
    sign = [token, timestamp, nonce] + list(args)
    sign.sort()
    sign = to_binary(''.join(sign))
    return sha1(sign).hexdigest()


def check_signature(token, timestamp, nonce, signature):
    if not (token and timestamp and nonce and signature):
        return False
    sign = get_signature(token, timestamp, nonce)
    return sign == signature


def check_token(token):
    return re.match('^[A-Za-z0-9]{3,32}$', token)


def cached_property(method):
    prop_name = '_{}'.format(method.__name__)

    @wraps(method)
    def wrapped_func(self, *args, **kwargs):
        if not hasattr(self, prop_name):
            setattr(self, prop_name, method(self, *args, **kwargs))
        return getattr(self, prop_name)

    return property(wrapped_func)


def to_text(value, encoding="utf-8") -> str:
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        return value.decode(encoding)
    return str(value)


def to_binary(value, encoding="utf-8") -> bytes:
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value.encode(encoding)
    return bytes(value)


def is_string(value) -> bool:
    """Check if value's type is `str` or `bytes`
    """
    return isinstance(value, string_types)


def byte2int(s, index=0):
    """Get the ASCII int value of a character in a string.

    :param s: a string
    :param index: the position of desired character

    :return: ASCII int value
    """
    return s[index]


def generate_token(length=''):
    if not length:
        length = random.randint(3, 32)
    length = int(length)
    assert 3 <= length <= 32
    letters = string.ascii_letters + string.digits
    return ''.join(choice(letters) for _ in range(length))


def json_loads(s):
    s = to_text(s)
    return json.loads(s)


def json_dumps(d):
    return json.dumps(d)


def pay_sign_dict(
    appid,
    pay_sign_key,
    add_noncestr=True,
    add_timestamp=True,
    add_appid=True,
    **kwargs
):
    """
    支付参数签名
    """
    assert pay_sign_key, "PAY SIGN KEY IS EMPTY"

    if add_appid:
        kwargs.update({'appid': appid})

    if add_noncestr:
        kwargs.update({'noncestr': generate_token()})

    if add_timestamp:
        kwargs.update({'timestamp': int(time.time())})

    params = kwargs.items()

    _params = [
        (k.lower(), v) for k, v in kwargs.items() if k.lower() != "appid"
    ]
    _params += [('appid', appid), ('appkey', pay_sign_key)]
    _params.sort()

    sign = '&'.join(["%s=%s" % (str(p[0]), str(p[1]))
                     for p in _params]).encode("utf-8")
    sign = sha1(sign).hexdigest()
    sign_type = 'SHA1'

    return dict(params), sign, sign_type


def make_error_page(url):
    with io.open(
        os.path.join(os.path.dirname(__file__), 'contrib/error.html'),
        'r',
        encoding='utf-8'
    ) as error_page:
        return error_page.read().replace('{url}', url)


def is_regex(value):
    return isinstance(value, re_type)
